from __future__ import annotations

import datetime

import pytest

import ibis
import ibis.common.exceptions as com
import ibis.expr.datatypes as dt
import ibis.expr.operations as ops
from ibis import selectors as s
from ibis.common.annotations import ValidationError
from ibis.common.deferred import _


def test_tumble_tvf_schema(schema, table):
    expr = table.window_by(time_col=table.i).tumble(
        window_size=ibis.interval(minutes=15)
    )
    expected_schema = ibis.schema(
        schema
        + [
            ("window_start", dt.Timestamp(scale=3)),
            ("window_end", dt.Timestamp(scale=3)),
            ("window_time", dt.Timestamp(scale=3)),
        ]
    )
    assert expr.schema() == expected_schema


@pytest.mark.parametrize("wrong_type_window_size", ["60", 60])
def test_create_tumble_tvf_with_wrong_scalar_type(table, wrong_type_window_size):
    with pytest.raises(ValidationError, match=".* is not coercible to a .*"):
        table.window_by(time_col=table.i).tumble(window_size=wrong_type_window_size)


def test_create_tumble_tvf_with_nonexistent_time_col(table):
    with pytest.raises(com.IbisTypeError, match="Column .* is not found in table"):
        table.window_by(time_col=table["nonexistent"]).tumble(
            window_size=datetime.timedelta(seconds=60)
        )


def test_create_tumble_tvf_with_nonscalar_window_size(schema):
    schema.append(("l", "interval"))
    table = ibis.table(schema, name="table")
    with pytest.raises(ValidationError, match=".* is not coercible to a .*"):
        table.window_by(time_col=table.i).tumble(window_size=table.l)


def test_create_tumble_tvf_with_non_timestamp_time_col(table):
    with pytest.raises(ValidationError, match=".* is not coercible to a .*"):
        table.window_by(time_col=table.e).tumble(window_size=ibis.interval(minutes=15))


def test_create_tumble_tvf_with_str_time_col(table):
    expr = table.window_by(time_col="i").tumble(window_size=ibis.interval(minutes=15))
    assert isinstance(expr.op(), ops.TumbleWindowingTVF)
    assert expr.op().time_col == table.i.op()


@pytest.mark.parametrize("deferred", [_["i"], _.i])
def test_create_tumble_tvf_with_deferred_time_col(table, deferred):
    expr = table.window_by(time_col=deferred.resolve(table)).tumble(
        window_size=ibis.interval(minutes=15)
    )
    assert isinstance(expr.op(), ops.TumbleWindowingTVF)
    assert expr.op().time_col == table.i.op()


def test_create_tumble_tvf_with_selector_time_col(table):
    expr = table.window_by(time_col=s.c("i")).tumble(
        window_size=ibis.interval(minutes=15)
    )
    assert isinstance(expr.op(), ops.TumbleWindowingTVF)
    assert expr.op().time_col == table.i.op()



from __future__ import annotations

import ibis

from .conftest import add_date, tpch_test


@tpch_test
def test_tpc_h12(orders, lineitem):
    """'Shipping Modes and Order Priority Query (Q12)

    This query determines whether selecting less expensive modes of shipping is
    negatively affecting the critical-prior- ity orders by causing more parts
    to be received by customers after the committed date."""
    SHIPMODE1 = "MAIL"
    SHIPMODE2 = "SHIP"
    DATE = "1994-01-01"

    q = orders
    q = q.join(lineitem, orders.o_orderkey == lineitem.l_orderkey)

    q = q.filter(
        [
            q.l_shipmode.isin([SHIPMODE1, SHIPMODE2]),
            q.l_commitdate < q.l_receiptdate,
            q.l_shipdate < q.l_commitdate,
            q.l_receiptdate >= ibis.date(DATE),
            q.l_receiptdate < add_date(DATE, dy=1),
        ]
    )

    gq = q.group_by([q.l_shipmode])
    q = gq.aggregate(
        high_line_count=(
            q.o_orderpriority.case()
            .when("1-URGENT", 1)
            .when("2-HIGH", 1)
            .else_(0)
            .end()
        ).sum(),
        low_line_count=(
            q.o_orderpriority.case()
            .when("1-URGENT", 0)
            .when("2-HIGH", 0)
            .else_(1)
            .end()
        ).sum(),
    )
    q = q.order_by(q.l_shipmode)

    return q



"""Test ibis.util utilities."""
from __future__ import annotations

import pytest

from ibis.util import PseudoHashable, flatten_iterable, import_object


@pytest.mark.parametrize(
    ("case", "expected"),
    [
        ([], []),
        ([1], [1]),
        ([1, 2], [1, 2]),
        ([[]], []),
        ([[1]], [1]),
        ([[1, 2]], [1, 2]),
        ([[1, [2, [3, [4]]]]], [1, 2, 3, 4]),
        ([[4, [3, [2, [1]]]]], [4, 3, 2, 1]),
        ([[[[4], 3], 2], 1], [4, 3, 2, 1]),
        ([[[[1], 2], 3], 4], [1, 2, 3, 4]),
        ([{(1,), frozenset({(2,)})}], {1, 2}),
        ({(1, (2,)): None, "a": None}, [1, 2, "a"]),
        (([x] for x in range(5)), list(range(5))),
        ({(1, (2, frozenset({(3,)})))}, [1, 2, 3]),
    ],
)
def test_flatten(case, expected):
    assert type(expected)(flatten_iterable(case)) == expected


@pytest.mark.parametrize("case", [1, "abc", b"abc", 2.0, object()])
def test_flatten_invalid_input(case):
    flat = flatten_iterable(case)

    with pytest.raises(TypeError):
        list(flat)


def test_import_object():
    import collections

    assert import_object("collections.defaultdict") is collections.defaultdict
    assert import_object("collections.abc.Mapping") is collections.abc.Mapping

    with pytest.raises(ImportError):
        import_object("this_module_probably.doesnt_exist")

    with pytest.raises(ImportError):
        import_object("collections.this_attribute_doesnt_exist")


# TODO(kszucs): add tests for promote_list and promote_tuple


def test_pseudo_hashable():
    class Unhashable:
        def __init__(self, value):
            self.value = value

        def __eq__(self, other):
            return isinstance(other, Unhashable) and self.value == other.value

    class MyList(list):
        pass

    class MyMap(dict):
        pass

    for obj in [1, "a", b"a", 2.0, object(), (), frozenset()]:
        with pytest.raises(TypeError, match="Cannot wrap a hashable object"):
            PseudoHashable(obj)

    # test unhashable sequences
    lst1 = [1, 2, 3]
    lst2 = [1, 2, 4]
    lst3 = MyList([1, 2, 3])
    ph1 = PseudoHashable(lst1)
    ph2 = PseudoHashable(lst2)
    ph3 = PseudoHashable(lst3)
    ph4 = PseudoHashable(lst1.copy())

    assert hash(ph1) == hash(ph1)
    assert hash(ph1) != hash(ph2)
    assert hash(ph1) != hash(ph3)
    assert hash(ph2) != hash(ph3)
    assert hash(ph3) == hash(ph3)
    assert hash(ph1) == hash(ph4)
    assert ph1 == ph1
    assert ph1 != ph2
    assert ph1 == ph3
    assert ph2 != ph3
    assert ph3 == ph3
    assert ph1 == ph4

    # test unhashable mappings
    dct1 = {"a": 1, "b": 2}
    dct2 = {"a": 1, "b": 3}
    dct3 = MyMap({"a": 1, "b": 2})
    ph1 = PseudoHashable(dct1)
    ph2 = PseudoHashable(dct2)
    ph3 = PseudoHashable(dct3)
    ph4 = PseudoHashable(dct1.copy())

    assert hash(ph1) == hash(ph1)
    assert hash(ph1) != hash(ph2)
    assert hash(ph1) != hash(ph3)
    assert hash(ph2) != hash(ph3)
    assert hash(ph3) == hash(ph3)
    assert hash(ph1) == hash(ph4)
    assert ph1 == ph1
    assert ph1 != ph2
    assert ph1 == ph3
    assert ph2 != ph3
    assert ph3 == ph3
    assert ph1 == ph4

    # test unhashable objects
    obj1 = Unhashable(1)
    obj2 = Unhashable(1)
    obj3 = Unhashable(2)
    obj4 = Unhashable(1)
    ph1 = PseudoHashable(obj1)
    ph2 = PseudoHashable(obj2)
    ph3 = PseudoHashable(obj3)
    ph4 = PseudoHashable(obj4)

    assert hash(ph1) == hash(ph1)
    assert hash(ph1) != hash(ph2)
    assert hash(ph1) != hash(ph3)
    assert hash(ph2) != hash(ph3)
    assert hash(ph3) == hash(ph3)
    assert hash(ph1) != hash(ph4)
    assert ph1 == ph1
    assert ph1 == ph2
    assert ph1 != ph3
    assert ph2 != ph3
    assert ph3 == ph3
    assert ph1 == ph4



from __future__ import annotations

import pytest

import ibis
from ibis.tests.expr.mocks import MockBackend


@pytest.fixture(scope="module")
def con():
    return MockBackend()


@pytest.fixture(scope="module")
def alltypes(con):
    return con.table("alltypes")


@pytest.fixture(scope="module")
def functional_alltypes(con):
    return con.table("functional_alltypes")


@pytest.fixture(scope="module")
def star1(con):
    return con.table("star1")


@pytest.fixture(scope="module")
def star2(con):
    return con.table("star2")


@pytest.fixture(scope="module")
def star3(con):
    return con.table("star3")


@pytest.fixture(scope="module")
def nation(con):
    return con.table("tpch_nation")


@pytest.fixture(scope="module")
def region(con):
    return con.table("tpch_region")


@pytest.fixture(scope="module")
def customer(con):
    return con.table("tpch_customer")


@pytest.fixture(scope="module")
def airlines(con):
    return con.table("airlines")


@pytest.fixture(scope="module")
def foo_t(con):
    return con.table("foo_t")


@pytest.fixture(scope="module")
def bar_t(con):
    return con.table("bar_t")


def to_sql(expr, *args, **kwargs) -> str:
    pytest.importorskip("duckdb")
    return str(ibis.to_sql(expr, *args, dialect="duckdb", **kwargs))


@pytest.fixture(scope="module")
def foo(con):
    return con.table("foo")


@pytest.fixture(scope="module")
def bar(con):
    return con.table("bar")


@pytest.fixture(scope="module")
def t1(con):
    return con.table("t1")


@pytest.fixture(scope="module")
def t2(con):
    return con.table("t2")


@pytest.fixture(scope="module")
def where_uncorrelated_subquery(foo, bar):
    return foo[foo.job.isin(bar.job)]


@pytest.fixture(scope="module")
def not_exists(foo_t, bar_t):
    return foo_t[-(foo_t.key1 == bar_t.key1).any()]


@pytest.fixture(scope="module")
def union(con):
    table = con.table("functional_alltypes")

    t1 = table[table.int_col > 0][
        table.string_col.name("key"),
        table.float_col.cast("double").name("value"),
    ]
    t2 = table[table.int_col <= 0][
        table.string_col.name("key"), table.double_col.name("value")
    ]

    return t1.union(t2, distinct=True)


@pytest.fixture(scope="module")
def union_all(con):
    table = con.table("functional_alltypes")

    t1 = table[table.int_col > 0][
        table.string_col.name("key"),
        table.float_col.cast("double").name("value"),
    ]
    t2 = table[table.int_col <= 0][
        table.string_col.name("key"), table.double_col.name("value")
    ]

    return t1.union(t2, distinct=False)


@pytest.fixture(scope="module")
def intersect(con):
    table = con.table("functional_alltypes")

    t1 = table[table.int_col > 0][
        table.string_col.name("key"),
        table.float_col.cast("double").name("value"),
    ]
    t2 = table[table.int_col <= 0][
        table.string_col.name("key"), table.double_col.name("value")
    ]

    return t1.intersect(t2)


@pytest.fixture(scope="module")
def difference(con):
    table = con.table("functional_alltypes")

    t1 = table[table.int_col > 0][
        table.string_col.name("key"),
        table.float_col.cast("double").name("value"),
    ]
    t2 = table[table.int_col <= 0][
        table.string_col.name("key"), table.double_col.name("value")
    ]

    return t1.difference(t2)


@pytest.fixture(scope="module")
def simple_case(con):
    t = con.table("alltypes")
    return t.g.case().when("foo", "bar").when("baz", "qux").else_("default").end()


@pytest.fixture(scope="module")
def search_case(con):
    t = con.table("alltypes")
    return ibis.case().when(t.f > 0, t.d * 2).when(t.c < 0, t.a * 2).end()


@pytest.fixture(scope="module")
def projection_fuse_filter():
    # Probably test this during the evaluation phase. In SQL, "fusable"
    # table operations will be combined together into a single select
    # statement
    #
    # see ibis #71 for more on this

    t = ibis.table(
        [
            ("a", "int8"),
            ("b", "int16"),
            ("c", "int32"),
            ("d", "int64"),
            ("e", "float32"),
            ("f", "float64"),
            ("g", "string"),
            ("h", "boolean"),
        ],
        "foo",
    )

    proj = t["a", "b", "c"]

    # Rewrite a little more aggressively here
    expr1 = proj[t.a > 0]

    # at one point these yielded different results
    filtered = t[t.a > 0]

    expr2 = filtered[t.a, t.b, t.c]
    expr3 = filtered.select(["a", "b", "c"])

    return expr1, expr2, expr3


@pytest.fixture(scope="module")
def startswith(star1):
    t1 = star1
    return t1.foo_id.startswith("foo")


@pytest.fixture(scope="module")
def endswith(star1):
    t1 = star1
    return t1.foo_id.endswith("foo")



from __future__ import annotations

import pytest
from pytest import param

import ibis.expr.datatypes as dt
from ibis.backends.base.sqlglot.datatypes import TrinoType

dtypes = [
    ("interval year to month", dt.Interval(unit="M")),
    ("interval day to second", dt.Interval(unit="ms")),
    ("bigint", dt.int64),
    ("boolean", dt.boolean),
    ("varbinary", dt.binary),
    ("double", dt.float64),
    ("real", dt.float32),
    ("smallint", dt.int16),
    ("timestamp", dt.Timestamp(scale=3)),
    ("timestamp(6)", dt.Timestamp(scale=6)),
    ("timestamp(9) with time zone", dt.Timestamp(scale=9, timezone="UTC")),
    ("timestamp with time zone", dt.Timestamp(scale=3, timezone="UTC")),
    ("date", dt.date),
    ("time", dt.time),
    ("time(6)", dt.time),
    ("time with time zone", dt.time),
    ("time(7) with time zone", dt.time),
    ("tinyint", dt.int8),
    ("integer", dt.int32),
    ("uuid", dt.uuid),
    ("char", dt.string),
    ("char(42)", dt.string),
    ("json", dt.json),
    ("ipaddress", dt.inet),
    ("varchar", dt.string),
    ("varchar(7)", dt.string),
    ("decimal", dt.Decimal(18, 3)),
    ("decimal(15, 0)", dt.Decimal(15, 0)),
    ("decimal(23, 5)", dt.Decimal(23, 5)),
    ("numeric", dt.Decimal(18, 3)),
    ("numeric(15, 0)", dt.Decimal(15, 0)),
    ("numeric(23, 5)", dt.Decimal(23, 5)),
    ("array(date)", dt.Array(dt.date)),
    ("array(array(date))", dt.Array(dt.Array(dt.date))),
    ("array(array(decimal(42, 23)))", dt.Array(dt.Array(dt.Decimal(42, 23)))),
    (
        "array(row(xYz map(varchar(3), double)))",
        dt.Array(dt.Struct(dict(xYz=dt.Map(dt.string, dt.float64)))),
    ),
    ("map(varchar, array(double))", dt.Map(dt.string, dt.Array(dt.float64))),
    (
        "row(a varchar, b array(tinyint), c map(bigint, row(d double)))",
        dt.Struct(
            dict(
                a=dt.string,
                b=dt.Array(dt.int8),
                c=dt.Map(dt.int64, dt.Struct(dict(d=dt.float64))),
            )
        ),
    ),
]


@pytest.mark.parametrize(
    ("trino_type", "ibis_type"),
    [param(trino_type, ibis_type, id=trino_type) for trino_type, ibis_type in dtypes],
)
def test_parse(trino_type, ibis_type):
    assert TrinoType.from_string(trino_type) == ibis_type



from __future__ import annotations

import datetime

from ibis.formats.pandas import PandasData


class MySQLPandasData(PandasData):
    # TODO(kszucs): this could be reused at other backends, like pyspark
    @classmethod
    def convert_Time(cls, s, dtype, pandas_type):
        def convert(timedelta):
            comps = timedelta.components
            return datetime.time(
                hour=comps.hours,
                minute=comps.minutes,
                second=comps.seconds,
                microsecond=comps.milliseconds * 1000 + comps.microseconds,
            )

        return s.map(convert, na_action="ignore")

    @classmethod
    def convert_Timestamp(cls, s, dtype, pandas_type):
        if s.dtype == "object":
            s = s.replace("0000-00-00 00:00:00", None)
        return super().convert_Timestamp(s, dtype, pandas_type)



from __future__ import annotations

import hypothesis as h
import hypothesis.extra.numpy as npst
import hypothesis.strategies as st
import numpy as np
import pytest
from packaging.version import parse as vparse

import ibis.expr.datatypes as dt
import ibis.tests.strategies as ibst
from ibis.formats.numpy import NumpySchema, NumpyType

roundtripable_types = st.deferred(
    lambda: (
        npst.integer_dtypes(endianness="=")
        | npst.floating_dtypes(endianness="=")
        | npst.datetime64_dtypes(max_period="ns", endianness="=")
        | npst.timedelta64_dtypes(max_period="s", endianness="=")
    )
)


@st.composite
def numpy_schema(draw, item_strategy=roundtripable_types, max_size=10):
    num_fields = draw(st.integers(min_value=0, max_value=max_size))
    names = draw(
        st.lists(st.text(), min_size=num_fields, max_size=num_fields, unique=True)
    )
    types = draw(st.lists(item_strategy, min_size=num_fields, max_size=num_fields))
    return list(zip(names, types))


def assert_dtype_roundtrip(
    numpy_type, ibis_type=None, restored_type=None, nullable=True
):
    dtype = NumpyType.to_ibis(numpy_type, nullable=nullable)
    if ibis_type is not None:
        assert dtype == ibis_type

    nptyp = NumpyType.from_ibis(dtype)
    if restored_type is None:
        restored_type = numpy_type
    assert nptyp == restored_type


@h.given(roundtripable_types)
def test_roundtripable_types(numpy_type):
    assert_dtype_roundtrip(numpy_type, nullable=False)
    assert_dtype_roundtrip(numpy_type, nullable=True)


@h.given(npst.unicode_string_dtypes(endianness="="))
def test_non_roundtripable_str_type(numpy_type):
    assert_dtype_roundtrip(
        numpy_type, dt.String(nullable=False), np.dtype("object"), nullable=False
    )
    assert_dtype_roundtrip(
        numpy_type, dt.String(nullable=True), np.dtype("object"), nullable=True
    )


@h.given(npst.byte_string_dtypes(endianness="="))
def test_non_roundtripable_bytes_type(numpy_type):
    assert_dtype_roundtrip(
        numpy_type, dt.Binary(nullable=False), np.dtype("object"), nullable=False
    )
    assert_dtype_roundtrip(
        numpy_type, dt.Binary(nullable=True), np.dtype("object"), nullable=True
    )


@h.given(
    ibst.null_dtype
    | ibst.variadic_dtypes()
    | ibst.decimal_dtypes()
    | ibst.struct_dtypes()
)
def test_variadic_to_numpy(ibis_type):
    assert NumpyType.from_ibis(ibis_type) == np.dtype("object")


@h.given(ibst.timestamp_dtype())
def test_timestamp_to_numpy(ibis_type):
    assert NumpyType.from_ibis(ibis_type) == np.dtype("datetime64[ns]")


@h.given(ibst.date_dtype())
def test_date_to_numpy(ibis_type):
    assert NumpyType.from_ibis(ibis_type) == np.dtype("datetime64[D]")


@h.given(ibst.time_dtype())
def test_time_to_numpy(ibis_type):
    assert NumpyType.from_ibis(ibis_type) == np.dtype("timedelta64[ns]")


@h.given(ibst.schema())
def test_schema_to_numpy(ibis_schema):
    numpy_schema = NumpySchema.from_ibis(ibis_schema)
    assert len(numpy_schema) == len(ibis_schema)

    for name, numpy_type in numpy_schema:
        assert numpy_type == NumpyType.from_ibis(ibis_schema[name])


@h.given(numpy_schema())
def test_schema_from_numpy(numpy_schema):
    ibis_schema = NumpySchema.to_ibis(numpy_schema)
    assert len(numpy_schema) == len(ibis_schema)

    for name, numpy_type in numpy_schema:
        assert NumpyType.from_ibis(ibis_schema[name]) == numpy_type


@pytest.mark.parametrize(
    ("numpy_dtype", "ibis_dtype"),
    [
        (np.bool_, dt.boolean),
        (np.int8, dt.int8),
        (np.int16, dt.int16),
        (np.int32, dt.int32),
        (np.int64, dt.int64),
        (np.uint8, dt.uint8),
        (np.uint16, dt.uint16),
        (np.uint32, dt.uint32),
        (np.uint64, dt.uint64),
        (np.float16, dt.float16),
        (np.float32, dt.float32),
        (np.float64, dt.float64),
        (np.double, dt.double),
        (np.str_, dt.string),
        (np.datetime64, dt.timestamp),
    ],
)
def test_dtype_from_numpy(numpy_dtype, ibis_dtype):
    assert NumpyType.to_ibis(np.dtype(numpy_dtype)) == ibis_dtype


def test_dtype_from_numpy_dtype_timedelta():
    if vparse(pytest.importorskip("pyarrow").__version__) < vparse("9"):
        pytest.skip("pyarrow < 9 globally mutates the timedelta64 numpy dtype")

    assert NumpyType.to_ibis(np.dtype(np.timedelta64)) == dt.Interval(unit="s")



from __future__ import annotations

import ibis
import ibis.expr.datatypes as dt
import ibis.expr.operations as ops


def test_sortkey_propagates_dtype_and_shape():
    k = ops.SortKey(ibis.literal(1), ascending=True)
    assert k.dtype == dt.int8
    assert k.shape.is_scalar()

    t = ibis.table([("a", "int16")], name="t")
    k = ops.SortKey(t.a, ascending=True)
    assert k.dtype == dt.int16
    assert k.shape.is_columnar()



from __future__ import annotations

import ibis

from .conftest import add_date, tpch_test


@tpch_test
def test_tpc_h08(part, supplier, region, lineitem, orders, customer, nation):
    """National Market Share Query (Q8)"""
    NATION = "BRAZIL"
    REGION = "AMERICA"
    TYPE = "ECONOMY ANODIZED STEEL"
    DATE = "1995-01-01"

    n1 = nation
    n2 = n1.view()

    q = part
    q = q.join(lineitem, part.p_partkey == lineitem.l_partkey)
    q = q.join(supplier, supplier.s_suppkey == lineitem.l_suppkey)
    q = q.join(orders, lineitem.l_orderkey == orders.o_orderkey)
    q = q.join(customer, orders.o_custkey == customer.c_custkey)
    q = q.join(n1, customer.c_nationkey == n1.n_nationkey)
    q = q.join(region, n1.n_regionkey == region.r_regionkey)
    q = q.join(n2, supplier.s_nationkey == n2.n_nationkey)

    q = q[
        orders.o_orderdate.year().name("o_year"),
        (lineitem.l_extendedprice * (1 - lineitem.l_discount)).name("volume"),
        n2.n_name.name("nation"),
        region.r_name,
        orders.o_orderdate,
        part.p_type,
    ]

    q = q.filter(
        [
            q.r_name == REGION,
            q.o_orderdate.between(ibis.date(DATE), add_date(DATE, dy=2, dd=-1)),
            q.p_type == TYPE,
        ]
    )

    q = q.mutate(
        nation_volume=ibis.case().when(q.nation == NATION, q.volume).else_(0).end()
    )
    gq = q.group_by([q.o_year])
    q = gq.aggregate(mkt_share=q.nation_volume.sum() / q.volume.sum())
    q = q.order_by([q.o_year])
    return q



from __future__ import annotations

from .conftest import tpch_test


@tpch_test
def test_tpc_h22(customer, orders):
    """Global Sales Opportunity Query (Q22)

    The Global Sales Opportunity Query identifies geographies where there are
    customers who may be likely to make a purchase."""

    COUNTRY_CODES = ("13", "31", "23", "29", "30", "18", "17")

    q = customer.filter(
        [
            customer.c_acctbal > 0.00,
            customer.c_phone.substr(0, 2).isin(COUNTRY_CODES),
        ]
    )
    q = q.aggregate(avg_bal=customer.c_acctbal.mean())

    custsale = customer.filter(
        [
            customer.c_phone.substr(0, 2).isin(COUNTRY_CODES),
            customer.c_acctbal > q.avg_bal,
            ~(orders.o_custkey == customer.c_custkey).any(),
        ]
    )
    custsale = custsale[
        customer.c_phone.substr(0, 2).name("cntrycode"), customer.c_acctbal
    ]

    gq = custsale.group_by(custsale.cntrycode)
    outerq = gq.aggregate(numcust=custsale.count(), totacctbal=custsale.c_acctbal.sum())

    return outerq.order_by(outerq.cntrycode)



import ibis


call = ibis.table(
    name="call",
    schema={
        "start_time": "timestamp",
        "end_time": "timestamp",
        "employee_id": "int64",
        "call_outcome_id": "int64",
        "call_attempts": "int64",
    },
)

result = call.aggregate([call.call_attempts.mean().name("mean")])



from __future__ import annotations

import pytest

import ibis
import ibis.expr.datatypes as dt
import ibis.expr.operations as ops
from ibis.common.deferred import _, const, deferred, var
from ibis.common.patterns import AnyOf, Check, pattern, replace
from ibis.util import Namespace

p = Namespace(pattern, module=ops)
d = Namespace(deferred, module=ops)

x = var("x")
y = var("y")
z = var("z")
params = var("params")

zero = ibis.literal(0)
one = ibis.literal(1)
two = ibis.literal(2)
three = ibis.literal(3)
add_1_2 = one + two

param_a = ibis.param("int8")
param_b = ibis.param("int8")
param_values = {param_a.op(): 1, param_b.op(): 2}

get_literal_value = p.Literal >> _.value
inc_integer_literal = p.Literal(x, dtype=dt.Integer) >> _.copy(value=x + 1)
sub_param_from_const = p.ScalarParameter >> d.Literal(
    const(param_values)[_], dtype=_.dtype
)


@replace(p.Add(p.Literal(x), p.Literal(y)))
def fold_literal_add(_, x, y):
    return ibis.literal(x + y).op()


simplifications = AnyOf(
    p.Add(x, p.Literal(0)) >> x,
    p.Add(p.Literal(0), x) >> x,
    p.Subtract(x, p.Literal(0)) >> x,
    p.Multiply(x, p.Literal(1)) >> x,
    p.Multiply(p.Literal(1), x) >> x,
    p.Multiply(x, p.Literal(0)) >> zero.op(),
    p.Multiply(p.Literal(0), x) >> zero.op(),
    p.Divide(x, p.Literal(1)) >> x,
    p.Divide(p.Literal(0), x) >> zero.op(),
    p.Divide(x, y) & Check(x == y) >> one.op(),
    fold_literal_add,
)


@replace(p.Literal(value=x, dtype=y))
def literal_to_type_call(_, x, y):
    return f"{y}({x})"


@pytest.mark.parametrize(
    ("rule", "expr", "expected"),
    [
        (get_literal_value, one, 1),
        (get_literal_value, two, 2),
        (literal_to_type_call, one, "int8(1)"),
        (inc_integer_literal, one, two.op()),
        (sub_param_from_const, param_a + param_b, add_1_2.op()),
    ],
)
def test_replace_scalar_parameters(rule, expr, expected):
    assert expr.op().replace(rule) == expected


def test_replace_scalar_parameters_using_variable():
    expr = param_a + param_b
    context = {"params": param_values}
    sub_param_from_var = p.ScalarParameter(x) >> d.Literal(params[_], dtype=x)
    assert expr.op().replace(sub_param_from_var, context=context) == add_1_2.op()


def test_replace_propagation():
    expr = add_1_2 + add_1_2 + add_1_2
    rule = p.Add(p.Add(x, y), z) >> d.Subtract(d.Subtract(x, y), z)
    result = expr.op().replace(rule)
    expected = ((one - two) - add_1_2) + add_1_2
    assert result == expected.op()


@pytest.mark.parametrize(
    ("expr", "expected"),
    [
        (one + zero, one),
        (zero + one, one),
        (one - zero, one),
        (one * zero, zero),
        (zero * one, zero),
        ((one + one + one) * one, three),
        (one / one, one),
        (one / one / one, one),
        (one / (one / one), one),
        (one / (one / one) / one, one),
        (three / (one / one) / one, three),
        (three / three, one),
    ],
)
def test_simplification(expr, expected):
    result = expr.op().replace(simplifications)
    assert result == expected.op()



from __future__ import annotations

from collections.abc import Mapping
from functools import lru_cache
from pathlib import Path
from typing import TYPE_CHECKING, Any

import polars as pl

import ibis
import ibis.common.exceptions as com
import ibis.expr.operations as ops
import ibis.expr.schema as sch
import ibis.expr.types as ir
from ibis.backends.base import BaseBackend, Database, NoUrl
from ibis.backends.base.sqlglot.dialects import Polars
from ibis.backends.pandas.rewrites import (
    bind_unbound_table,
    replace_parameter,
    rewrite_join,
)
from ibis.backends.polars.compiler import translate
from ibis.backends.polars.datatypes import dtype_to_polars, schema_from_polars
from ibis.util import gen_name, normalize_filename

if TYPE_CHECKING:
    from collections.abc import Iterable

    import pandas as pd
    import pyarrow as pa


class Backend(BaseBackend, NoUrl):
    name = "polars"
    dialect = Polars

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._tables = dict()
        self._context = pl.SQLContext()

    def do_connect(
        self, tables: Mapping[str, pl.LazyFrame | pl.DataFrame] | None = None
    ) -> None:
        """Construct a client from a dictionary of polars `LazyFrame`s and/or `DataFrame`s.

        Parameters
        ----------
        tables
            An optional mapping of string table names to polars LazyFrames.

        """
        if tables is not None and not isinstance(tables, Mapping):
            raise TypeError("Input to ibis.polars.connect must be a mapping")

        # tables are emphemeral
        self._tables.clear()

        for name, table in (tables or {}).items():
            self._add_table(name, table)

    def disconnect(self) -> None:
        pass

    @property
    def version(self) -> str:
        return pl.__version__

    def list_tables(self, like=None, database=None):
        return self._filter_with_like(list(self._tables.keys()), like)

    def table(self, name: str, _schema: sch.Schema | None = None) -> ir.Table:
        schema = schema_from_polars(self._tables[name].schema)
        return ops.DatabaseTable(name, schema, self).to_expr()

    def register(
        self,
        source: str | Path | Any,
        table_name: str | None = None,
        **kwargs: Any,
    ) -> ir.Table:
        """Register a data source as a table in the current database.

        Parameters
        ----------
        source
            The data source(s). May be a path to a file, a parquet directory, or a pandas
            dataframe.
        table_name
            An optional name to use for the created table. This defaults to
            a sequentially generated name.
        **kwargs
            Additional keyword arguments passed to Polars loading functions for
            CSV or parquet.
            See https://pola-rs.github.io/polars/py-polars/html/reference/api/polars.scan_csv.html
            and https://pola-rs.github.io/polars/py-polars/html/reference/api/polars.scan_parquet.html
            for more information

        Returns
        -------
        ir.Table
            The just-registered table

        """

        if isinstance(source, (str, Path)):
            first = str(source)
        elif isinstance(source, (list, tuple)):
            first = str(source[0])
        else:
            try:
                return self.read_pandas(source, table_name=table_name, **kwargs)
            except ValueError:
                self._register_failure()

        if first.startswith(("parquet://", "parq://")) or first.endswith(
            ("parq", "parquet")
        ):
            return self.read_parquet(source, table_name=table_name, **kwargs)
        elif first.startswith(
            ("csv://", "csv.gz://", "txt://", "txt.gz://")
        ) or first.endswith(("csv", "csv.gz", "tsv", "tsv.gz", "txt", "txt.gz")):
            return self.read_csv(source, table_name=table_name, **kwargs)
        else:
            self._register_failure()
        return None

    def _register_failure(self):
        import inspect

        msg = ", ".join(
            m[0] for m in inspect.getmembers(self) if m[0].startswith("read_")
        )
        raise ValueError(
            f"Cannot infer appropriate read function for input, "
            f"please call one of {msg} directly"
        )

    def _add_table(self, name: str, obj: pl.LazyFrame | pl.DataFrame) -> None:
        if isinstance(obj, pl.DataFrame):
            obj = obj.lazy()
        self._tables[name] = obj
        self._context.register(name, obj)

    def _remove_table(self, name: str) -> None:
        del self._tables[name]
        self._context.unregister(name)

    def sql(
        self, query: str, schema: sch.Schema | None = None, dialect: str | None = None
    ) -> ir.Table:
        query = self._transpile_sql(query, dialect=dialect)
        if schema is None:
            schema = self._get_schema_using_query(query)
        name = ibis.util.gen_name("polars_dot_sql_table")
        self._add_table(name, self._context.execute(query))
        return self.table(name)

    def read_csv(
        self, path: str | Path, table_name: str | None = None, **kwargs: Any
    ) -> ir.Table:
        """Register a CSV file as a table.

        Parameters
        ----------
        path
            The data source. A string or Path to the CSV file.
        table_name
            An optional name to use for the created table. This defaults to
            a sequentially generated name.
        **kwargs
            Additional keyword arguments passed to Polars loading function.
            See https://pola-rs.github.io/polars/py-polars/html/reference/api/polars.scan_csv.html
            for more information.

        Returns
        -------
        ir.Table
            The just-registered table

        """
        path = normalize_filename(path)
        table_name = table_name or gen_name("read_csv")
        try:
            self._add_table(table_name, pl.scan_csv(path, **kwargs))
        except pl.exceptions.ComputeError:
            # handles compressed csvs
            self._add_table(table_name, pl.read_csv(path, **kwargs))
        return self.table(table_name)

    def read_json(
        self, path: str | Path, table_name: str | None = None, **kwargs: Any
    ) -> ir.Table:
        """Register a JSON file as a table.

        Parameters
        ----------
        path
            A string or Path to a JSON file; globs are supported
        table_name
            An optional name to use for the created table. This defaults to
            a sequentially generated name.
        **kwargs
            Additional keyword arguments passed to Polars loading function.
            See https://pola-rs.github.io/polars/py-polars/html/reference/api/polars.scan_ndjson.html
            for more information.

        Returns
        -------
        ir.Table
            The just-registered table

        """
        path = normalize_filename(path)
        table_name = table_name or gen_name("read_json")
        try:
            self._add_table(table_name, pl.scan_ndjson(path, **kwargs))
        except pl.exceptions.ComputeError:
            # handles compressed json files
            self._add_table(table_name, pl.read_ndjson(path, **kwargs))
        return self.table(table_name)

    def read_delta(
        self, path: str | Path, table_name: str | None = None, **kwargs: Any
    ) -> ir.Table:
        """Register a Delta Lake as a table in the current database.

        Parameters
        ----------
        path
            The data source(s). Path to a Delta Lake table directory.
        table_name
            An optional name to use for the created table. This defaults to
            a sequentially generated name.
        **kwargs
            Additional keyword arguments passed to Polars loading function.
            See https://pola-rs.github.io/polars/py-polars/html/reference/api/polars.scan_delta.html
            for more information.

        Returns
        -------
        ir.Table
            The just-registered table

        """
        try:
            import deltalake  # noqa: F401
        except ImportError:
            raise ImportError(
                "The deltalake extra is required to use the "
                "read_delta method. You can install it using pip:\n\n"
                "pip install 'ibis-framework[polars,deltalake]'\n"
            )
        path = normalize_filename(path)
        table_name = table_name or gen_name("read_delta")
        self._add_table(table_name, pl.scan_delta(path, **kwargs))
        return self.table(table_name)

    def read_pandas(
        self, source: pd.DataFrame, table_name: str | None = None, **kwargs: Any
    ) -> ir.Table:
        """Register a Pandas DataFrame or pyarrow Table a table in the current database.

        Parameters
        ----------
        source
            The data source.
        table_name
            An optional name to use for the created table. This defaults to
            a sequentially generated name.
        **kwargs
            Additional keyword arguments passed to Polars loading function.
            See https://pola-rs.github.io/polars/py-polars/html/reference/api/polars.from_pandas.html
            for more information.

        Returns
        -------
        ir.Table
            The just-registered table

        """
        table_name = table_name or gen_name("read_in_memory")
        self._add_table(table_name, pl.from_pandas(source, **kwargs).lazy())
        return self.table(table_name)

    def read_parquet(
        self,
        path: str | Path | Iterable[str],
        table_name: str | None = None,
        **kwargs: Any,
    ) -> ir.Table:
        """Register a parquet file as a table in the current database.

        Parameters
        ----------
        path
            The data source(s). May be a path to a file, an iterable of files,
            or directory of parquet files.
        table_name
            An optional name to use for the created table. This defaults to
            a sequentially generated name.
        **kwargs
            Additional keyword arguments passed to Polars loading function.
            See https://pola-rs.github.io/polars/py-polars/html/reference/api/polars.scan_parquet.html
            for more information (if loading a single file or glob; when loading
            multiple files polars' `scan_pyarrow_dataset` method is used instead).

        Returns
        -------
        ir.Table
            The just-registered table

        """
        table_name = table_name or gen_name("read_parquet")
        if not isinstance(path, (str, Path)) and len(path) == 1:
            path = path[0]

        if not isinstance(path, (str, Path)) and len(path) > 1:
            self._import_pyarrow()
            import pyarrow.dataset as ds

            paths = [normalize_filename(p) for p in path]
            obj = pl.scan_pyarrow_dataset(
                source=ds.dataset(paths, format="parquet"),
                **kwargs,
            )
            self._add_table(table_name, obj)
        else:
            path = normalize_filename(path)
            self._add_table(table_name, pl.scan_parquet(path, **kwargs))

        return self.table(table_name)

    def database(self, name=None):
        return Database(name, self)

    def create_table(
        self,
        name: str,
        obj: pd.DataFrame | pa.Table | ir.Table | None = None,
        *,
        schema: ibis.Schema | None = None,
        database: str | None = None,
        temp: bool | None = None,
        overwrite: bool = False,
    ) -> ir.Table:
        if schema is not None and obj is None:
            obj = pl.LazyFrame(
                [],
                schema={name: dtype_to_polars(dtype) for name, dtype in schema.items()},
            )

        if database is not None:
            raise com.IbisError(
                "Passing `database` to the Polars backend create_table method has no "
                "effect: Polars cannot set a database."
            )

        if temp is False:
            raise com.IbisError(
                "Passing `temp=False` to the Polars backend create_table method is not "
                "supported: all tables are in memory and temporary."
            )

        if not overwrite and name in self._tables:
            raise com.IntegrityError(
                f"Table {name} already exists. Use overwrite=True to clobber existing tables"
            )

        if isinstance(obj, ir.Table):
            obj = self.to_pyarrow(obj)

        if not isinstance(obj, (pl.DataFrame, pl.LazyFrame)):
            obj = pl.LazyFrame(obj)

        self._add_table(name, obj)
        return self.table(name)

    def get_schema(self, table_name, database=None):
        return self._tables[table_name].schema

    @classmethod
    @lru_cache
    def _get_operations(cls):
        return tuple(op for op in translate.registry if issubclass(op, ops.Value))

    @classmethod
    def has_operation(cls, operation: type[ops.Value]) -> bool:
        # Polars doesn't support geospatial ops, but the dispatcher implements
        # a common base class that makes it appear that it does. Explicitly
        # exclude these operations.
        if issubclass(
            operation, (ops.GeoSpatialUnOp, ops.GeoSpatialBinOp, ops.GeoUnaryUnion)
        ):
            return False
        op_classes = cls._get_operations()
        return operation in op_classes or issubclass(operation, op_classes)

    def compile(
        self, expr: ir.Expr, params: Mapping[ir.Expr, object] | None = None, **_: Any
    ):
        if params is None:
            params = dict()
        else:
            params = {param.op(): value for param, value in params.items()}

        node = expr.as_table().op()
        node = node.replace(
            rewrite_join | replace_parameter | bind_unbound_table,
            context={"params": params, "backend": self},
        )

        return translate(node, ctx=self._context)

    def _get_sql_string_view_schema(self, name, table, query) -> sch.Schema:
        raise NotImplementedError("table.sql() not yet supported in polars")

    def _get_schema_using_query(self, query: str) -> sch.Schema:
        return schema_from_polars(self._context.execute(query).schema)

    def execute(
        self,
        expr: ir.Expr,
        params: Mapping[ir.Expr, object] | None = None,
        limit: int | None = None,
        streaming: bool = False,
        **kwargs: Any,
    ):
        lf = self.compile(expr, params=params, **kwargs)
        if limit == "default":
            limit = ibis.options.sql.default_limit
        if limit is not None:
            df = lf.fetch(limit, streaming=streaming)
        else:
            df = lf.collect(streaming=streaming)

        if isinstance(expr, (ir.Table, ir.Scalar)):
            return expr.__pandas_result__(df.to_pandas())
        else:
            assert isinstance(expr, ir.Column), type(expr)
            if expr.type().is_temporal():
                return expr.__pandas_result__(df.to_pandas())
            else:
                # note: skip frame-construction overhead
                return df.to_series().to_pandas()

    def _to_pyarrow_table(
        self,
        expr: ir.Expr,
        params: Mapping[ir.Expr, object] | None = None,
        limit: int | None = None,
        streaming: bool = False,
        **kwargs: Any,
    ):
        lf = self.compile(expr, params=params, **kwargs)
        if limit is not None:
            df = lf.fetch(limit, streaming=streaming)
        else:
            df = lf.collect(streaming=streaming)

        table = df.to_arrow()
        if isinstance(expr, (ir.Table, ir.Value)):
            schema = expr.as_table().schema().to_pyarrow()
            return table.rename_columns(schema.names).cast(schema)
        else:
            raise com.IbisError(f"Cannot execute expression of type: {type(expr)}")

    def to_pyarrow(
        self,
        expr: ir.Expr,
        params: Mapping[ir.Expr, object] | None = None,
        limit: int | None = None,
        **kwargs: Any,
    ):
        result = self._to_pyarrow_table(expr, params=params, limit=limit, **kwargs)
        return expr.__pyarrow_result__(result)

    def to_pyarrow_batches(
        self,
        expr: ir.Expr,
        *,
        params: Mapping[ir.Scalar, Any] | None = None,
        limit: int | str | None = None,
        chunk_size: int = 1_000_000,
        **kwargs: Any,
    ):
        self._import_pyarrow()
        table = self._to_pyarrow_table(expr, params=params, limit=limit, **kwargs)
        return table.to_reader(chunk_size)

    def _load_into_cache(self, name, expr):
        self.create_table(name, self.compile(expr).cache())

    def _clean_up_cached_table(self, op):
        self._remove_table(op.name)

    def create_view(self, *_, **__) -> ir.Table:
        raise NotImplementedError(self.name)

    def drop_table(self, *_, **__) -> ir.Table:
        raise NotImplementedError(self.name)

    def drop_view(self, *_, **__) -> ir.Table:
        raise NotImplementedError(self.name)



from __future__ import annotations

import json

import sqlglot as sg

import ibis.expr.schema as sch
from ibis.backends.base.sqlglot.datatypes import ImpalaType
from ibis.backends.base.sqlglot.ddl import DDL, DML, CreateDDL, DropFunction, DropObject


class ImpalaBase:
    dialect = "hive"

    def sanitize_format(self, format):
        _format_aliases = {"TEXT": "TEXTFILE"}

        if format is None:
            return None
        format = format.upper()
        format = _format_aliases.get(format, format)
        if format not in ("PARQUET", "AVRO", "TEXTFILE"):
            raise ValueError(f"Invalid format: {format!r}")

        return format

    def format_dtype(self, dtype):
        return ImpalaType.to_string(dtype)

    def format_properties(self, props):
        tokens = []
        for k, v in sorted(props.items()):
            tokens.append(f"  '{k}'='{v}'")
        return "(\n{}\n)".format(",\n".join(tokens))

    def format_tblproperties(self, props):
        formatted_props = self.format_properties(props)
        return f"TBLPROPERTIES {formatted_props}"

    def format_serdeproperties(self, props):
        formatted_props = self.format_properties(props)
        return f"SERDEPROPERTIES {formatted_props}"


class CreateDatabase(ImpalaBase, CreateDDL):
    def __init__(self, name, path=None, can_exist=False):
        self.name = name
        self.path = path
        self.can_exist = can_exist

    def compile(self):
        name = self.quote(self.name)

        create_decl = "CREATE DATABASE"
        create_line = f"{create_decl} {self._if_exists()}{name}"
        if self.path is not None:
            create_line += f"\nLOCATION '{self.path}'"

        return create_line


class DropDatabase(ImpalaBase, DropObject):
    _object_type = "DATABASE"

    def __init__(self, name, must_exist=True):
        super().__init__(must_exist=must_exist)
        self.name = name

    def _object_name(self):
        return self.name


class CreateTable(ImpalaBase, CreateDDL):
    def __init__(
        self,
        table_name,
        database=None,
        external=False,
        format="parquet",
        can_exist=False,
        partition=None,
        path=None,
        tbl_properties=None,
    ):
        self.table_name = table_name
        self.database = database
        self.partition = partition
        self.path = path
        self.external = external
        self.can_exist = can_exist
        self.format = self.sanitize_format(format)
        self.tbl_properties = tbl_properties

    @property
    def _prefix(self):
        if self.external:
            return "CREATE EXTERNAL TABLE"
        else:
            return "CREATE TABLE"

    def _create_line(self):
        scoped_name = self.scoped_name(self.table_name, self.database)
        return f"{self._prefix} {self._if_exists()}{scoped_name}"

    def _location(self):
        return f"LOCATION '{self.path}'" if self.path else None

    def _storage(self):
        # By the time we're here, we have a valid format
        return f"STORED AS {self.format}"

    @property
    def pieces(self):
        yield self._create_line()
        yield from filter(None, self._pieces)

    def compile(self):
        return "\n".join(self.pieces)


class CreateTableWithSchema(CreateTable):
    def __init__(self, table_name, schema, table_format=None, **kwargs):
        super().__init__(table_name, **kwargs)
        self.schema = schema
        self.table_format = table_format

    @property
    def _pieces(self):
        if self.partition is not None:
            main_schema = self.schema
            part_schema = self.partition
            if not isinstance(part_schema, sch.Schema):
                part_fields = {name: self.schema[name] for name in part_schema}
                part_schema = sch.Schema(part_fields)

            to_delete = {name for name in self.partition if name in self.schema}
            fields = {
                name: dtype
                for name, dtype in main_schema.items()
                if name not in to_delete
            }
            main_schema = sch.Schema(fields)

            yield self.format_schema(main_schema)
            yield f"PARTITIONED BY {self.format_schema(part_schema)}"
        else:
            yield self.format_schema(self.schema)

        if self.table_format is not None:
            yield "\n".join(self.table_format.to_ddl())
        else:
            yield self._storage()

        yield self._location()


class AlterTable(ImpalaBase, DDL):
    def __init__(
        self,
        table,
        location=None,
        format=None,
        tbl_properties=None,
        serde_properties=None,
    ):
        self.table = table
        self.location = location
        self.format = self.sanitize_format(format)
        self.tbl_properties = tbl_properties
        self.serde_properties = serde_properties

    def _wrap_command(self, cmd):
        return f"ALTER TABLE {cmd}"

    def _format_properties(self, prefix=""):
        tokens = []

        if self.location is not None:
            tokens.append(f"LOCATION '{self.location}'")

        if self.format is not None:
            tokens.append(f"FILEFORMAT {self.format}")

        if self.tbl_properties is not None:
            tokens.append(self.format_tblproperties(self.tbl_properties))

        if self.serde_properties is not None:
            tokens.append(self.format_serdeproperties(self.serde_properties))

        if len(tokens) > 0:
            return "\n{}{}".format(prefix, "\n".join(tokens))
        else:
            return ""

    def compile(self):
        props = self._format_properties()
        action = f"{self.table} SET {props}"
        return self._wrap_command(action)


class RenameTable(AlterTable):
    def __init__(
        self,
        old_name: str,
        new_name: str,
        old_database: str | None = None,
        new_database: str | None = None,
    ):
        self._old = sg.table(old_name, db=old_database, quoted=True).sql(
            dialect=self.dialect
        )
        self._new = sg.table(new_name, db=new_database, quoted=True).sql(
            dialect=self.dialect
        )

    def compile(self):
        return self._wrap_command(f"{self._old} RENAME TO {self._new}")


class DropTable(ImpalaBase, DropObject):
    _object_type = "TABLE"

    def __init__(self, table_name, database=None, must_exist=True):
        super().__init__(must_exist=must_exist)
        self.table_name = table_name
        self.database = database

    def _object_name(self):
        return self.scoped_name(self.table_name, self.database)


class TruncateTable(ImpalaBase, DDL):
    _object_type = "TABLE"

    def __init__(self, table_name, database=None):
        self.table_name = table_name
        self.database = database

    def compile(self):
        name = self.scoped_name(self.table_name, self.database)
        return f"TRUNCATE TABLE {name}"


class DropView(DropTable):
    _object_type = "VIEW"


class CTAS(CreateTable):
    """Create Table As Select."""

    def __init__(
        self,
        table_name,
        select,
        database=None,
        external=False,
        format="parquet",
        can_exist=False,
        path=None,
        partition=None,
    ):
        super().__init__(
            table_name,
            database=database,
            external=external,
            format=format,
            can_exist=can_exist,
            path=path,
            partition=partition,
        )
        self.select = select

    @property
    def _pieces(self):
        yield self._partitioned_by()
        yield self._storage()
        yield self._location()
        yield "AS"
        yield self.select

    def _partitioned_by(self):
        if self.partition is not None:
            return "PARTITIONED BY ({})".format(
                ", ".join(self.quote(expr.get_name()) for expr in self.partition)
            )
        return None


class CreateView(CTAS):
    """Create a view."""

    def __init__(self, table_name, select, database=None, can_exist=False):
        super().__init__(table_name, select, database=database, can_exist=can_exist)

    @property
    def _pieces(self):
        yield "AS"
        yield self.select

    @property
    def _prefix(self):
        return "CREATE VIEW"


class CreateTableParquet(CreateTable):
    def __init__(
        self,
        table_name,
        path,
        example_file=None,
        example_table=None,
        schema=None,
        external=True,
        **kwargs,
    ):
        super().__init__(
            table_name,
            external=external,
            format="parquet",
            path=path,
            **kwargs,
        )
        self.example_file = example_file
        self.example_table = example_table
        self.schema = schema

    @property
    def _pieces(self):
        if self.example_file is not None:
            yield f"LIKE PARQUET '{self.example_file}'"
        elif self.example_table is not None:
            yield f"LIKE {self.example_table}"
        elif self.schema is not None:
            yield self.format_schema(self.schema)
        else:
            raise NotImplementedError

        yield self._storage()
        yield self._location()


class DelimitedFormat(ImpalaBase):
    def __init__(
        self,
        path,
        delimiter=None,
        escapechar=None,
        na_rep=None,
        lineterminator=None,
    ):
        self.path = path
        self.delimiter = delimiter
        self.escapechar = escapechar
        self.lineterminator = lineterminator
        self.na_rep = na_rep

    def to_ddl(self):
        yield "ROW FORMAT DELIMITED"

        if self.delimiter is not None:
            yield f"FIELDS TERMINATED BY '{self.delimiter}'"

        if self.escapechar is not None:
            yield f"ESCAPED BY '{self.escapechar}'"

        if self.lineterminator is not None:
            yield f"LINES TERMINATED BY '{self.lineterminator}'"

        yield "STORED AS TEXTFILE"
        yield f"LOCATION '{self.path}'"

        if self.na_rep is not None:
            props = {"serialization.null.format": self.na_rep}
            yield self.format_tblproperties(props)


class AvroFormat(ImpalaBase):
    def __init__(self, path, avro_schema):
        self.path = path
        self.avro_schema = avro_schema

    def to_ddl(self):
        yield "STORED AS AVRO"
        yield f"LOCATION '{self.path}'"

        schema = json.dumps(self.avro_schema, indent=2, sort_keys=True)
        schema = "\n".join(x.rstrip() for x in schema.splitlines())

        props = {"avro.schema.literal": schema}
        yield self.format_tblproperties(props)


class ParquetFormat(ImpalaBase):
    def __init__(self, path):
        self.path = path

    def to_ddl(self):
        yield "STORED AS PARQUET"
        yield f"LOCATION '{self.path}'"


class CreateTableDelimited(CreateTableWithSchema):
    def __init__(
        self,
        table_name,
        path,
        schema,
        delimiter=None,
        escapechar=None,
        lineterminator=None,
        na_rep=None,
        external=True,
        **kwargs,
    ):
        table_format = DelimitedFormat(
            path,
            delimiter=delimiter,
            escapechar=escapechar,
            lineterminator=lineterminator,
            na_rep=na_rep,
        )
        super().__init__(table_name, schema, table_format, external=external, **kwargs)


class CreateTableAvro(CreateTable):
    def __init__(self, table_name, path, avro_schema, external=True, **kwargs):
        super().__init__(table_name, external=external, **kwargs)
        self.table_format = AvroFormat(path, avro_schema)

    @property
    def _pieces(self):
        yield "\n".join(self.table_format.to_ddl())


class LoadData(ImpalaBase, DDL):
    """Generate DDL for LOAD DATA command.

    Cannot be cancelled
    """

    def __init__(
        self,
        table_name,
        path,
        database=None,
        partition=None,
        partition_schema=None,
        overwrite=False,
    ):
        self.table_name = table_name
        self.database = database
        self.path = path

        self.partition = partition
        self.partition_schema = partition_schema

        self.overwrite = overwrite

    def compile(self):
        overwrite = "OVERWRITE " if self.overwrite else ""

        if self.partition is not None:
            partition = "\n" + self.format_partition(
                self.partition, self.partition_schema
            )
        else:
            partition = ""

        scoped_name = self.scoped_name(self.table_name, self.database)
        return "LOAD DATA INPATH '{}' {}INTO TABLE {}{}".format(
            self.path, overwrite, scoped_name, partition
        )


class PartitionProperties(AlterTable):
    def __init__(
        self,
        table,
        partition,
        partition_schema,
        location=None,
        format=None,
        tbl_properties=None,
        serde_properties=None,
    ):
        super().__init__(
            table,
            location=location,
            format=format,
            tbl_properties=tbl_properties,
            serde_properties=serde_properties,
        )
        self.partition = partition
        self.partition_schema = partition_schema

    def _compile(self, cmd, property_prefix=""):
        part = self.format_partition(self.partition, self.partition_schema)
        if cmd:
            part = f"{cmd} {part}"

        props = self._format_properties(property_prefix)
        action = f"{self.table} {part}{props}"
        return self._wrap_command(action)


class AddPartition(PartitionProperties):
    dialect = "hive"

    def __init__(self, table, partition, partition_schema, location=None):
        super().__init__(table, partition, partition_schema, location=location)

    def compile(self):
        return self._compile("ADD")


class AlterPartition(PartitionProperties):
    dialect = "hive"

    def compile(self):
        return self._compile("", "SET ")


class DropPartition(PartitionProperties):
    dialect = "hive"

    def __init__(self, table, partition, partition_schema):
        super().__init__(table, partition, partition_schema)

    def compile(self):
        return self._compile("DROP")


class CacheTable(ImpalaBase, DDL):
    def __init__(self, table_name, database=None, pool="default"):
        self.table_name = table_name
        self.database = database
        self.pool = pool

    def compile(self):
        scoped_name = self.scoped_name(self.table_name, self.database)
        return f"ALTER TABLE {scoped_name} SET CACHED IN '{self.pool}'"


class CreateFunction(ImpalaBase, DDL):
    _object_type = "FUNCTION"

    def __init__(self, func, name=None, database=None):
        self.func = func
        self.name = name or func.name
        self.database = database

    def _impala_signature(self):
        scoped_name = self.scoped_name(self.name, self.database)
        input_sig = ", ".join(map(self.format_dtype, self.func.inputs))
        output_sig = self.format_dtype(self.func.output)

        return f"{scoped_name}({input_sig}) returns {output_sig}"


class CreateUDF(CreateFunction):
    def compile(self):
        create_decl = "CREATE FUNCTION"
        impala_sig = self._impala_signature()
        param_line = f"location '{self.func.lib_path}' symbol='{self.func.so_symbol}'"
        return f"{create_decl} {impala_sig} {param_line}"


class CreateUDA(CreateFunction):
    def compile(self):
        create_decl = "CREATE AGGREGATE FUNCTION"
        impala_sig = self._impala_signature()
        tokens = [f"location '{self.func.lib_path}'"]

        fn_names = (
            "init_fn",
            "update_fn",
            "merge_fn",
            "serialize_fn",
            "finalize_fn",
        )

        for fn in fn_names:
            value = getattr(self.func, fn)
            if value is not None:
                tokens.append(f"{fn}='{value}'")

        joined_tokens = "\n".join(tokens)
        return f"{create_decl} {impala_sig} {joined_tokens}"


class DropFunction(ImpalaBase, DropFunction):
    def _impala_signature(self):
        full_name = self.scoped_name(self.name, self.database)
        input_sig = ", ".join(map(self.format_dtype, self.inputs))
        return f"{full_name}({input_sig})"


class ListFunction(ImpalaBase, DDL):
    def __init__(self, database, like=None, aggregate=False):
        self.database = database
        self.like = like
        self.aggregate = aggregate

    def compile(self):
        statement = "SHOW "
        if self.aggregate:
            statement += "AGGREGATE "
        statement += f"FUNCTIONS IN {self.database}"
        if self.like:
            statement += f" LIKE '{self.like}'"
        return statement


class InsertSelect(ImpalaBase, DML):
    def __init__(
        self,
        table_name,
        select_expr,
        database=None,
        partition=None,
        partition_schema=None,
        overwrite=False,
    ):
        self.table_name = table_name
        self.database = database
        self.select = select_expr

        self.partition = partition
        self.partition_schema = partition_schema

        self.overwrite = overwrite

    def compile(self):
        if self.overwrite:
            cmd = "INSERT OVERWRITE"
        else:
            cmd = "INSERT INTO"

        if self.partition is not None:
            part = self.format_partition(self.partition, self.partition_schema)
            partition = f" {part} "
        else:
            partition = ""

        select_query = self.select
        scoped_name = self.scoped_name(self.table_name, self.database)
        return f"{cmd} {scoped_name}{partition}\n{select_query}"



from __future__ import annotations

import os
import subprocess
from typing import TYPE_CHECKING

import sqlglot as sg

import ibis
from ibis.backends.tests.base import (
    ServiceBackendTest,
)

if TYPE_CHECKING:
    from collections.abc import Iterable
    from pathlib import Path
    from typing import Any

EXASOL_USER = os.environ.get("IBIS_TEST_EXASOL_USER", "sys")
EXASOL_PASS = os.environ.get("IBIS_TEST_EXASOL_PASSWORD", "exasol")
EXASOL_HOST = os.environ.get("IBIS_TEST_EXASOL_HOST", "localhost")
EXASOL_PORT = int(os.environ.get("IBIS_TEST_EXASOL_PORT", 8563))
IBIS_TEST_EXASOL_DB = os.environ.get("IBIS_TEST_EXASOL_DATABASE", "EXASOL")


class TestConf(ServiceBackendTest):
    check_dtype = False
    check_names = False
    supports_arrays = False
    supports_arrays_outside_of_select = supports_arrays
    supports_window_operations = True
    supports_divide_by_zero = False
    returned_timestamp_unit = "us"
    supported_to_timestamp_units = {"s", "ms", "us"}
    supports_floating_modulus = True
    native_bool = True
    supports_structs = False
    supports_json = False
    supports_map = False
    reduction_tolerance = 1e-7
    stateful = True
    service_name = "exasol"
    supports_tpch = False
    force_sort = True
    deps = ("pyexasol",)

    @staticmethod
    def connect(*, tmpdir, worker_id, **kw: Any):
        return ibis.exasol.connect(
            user=EXASOL_USER,
            password=EXASOL_PASS,
            host=EXASOL_HOST,
            port=EXASOL_PORT,
            **kw,
        )

    def postload(self, **kw: Any):
        self.connection = self.connect(schema=IBIS_TEST_EXASOL_DB, **kw)

    @staticmethod
    def format_table(name: str) -> str:
        return sg.to_identifier(name, quoted=True).sql("exasol")

    @property
    def test_files(self) -> Iterable[Path]:
        return self.data_dir.joinpath("csv").glob("*.csv")

    def _exaplus(self) -> str:
        find_exaplus = [
            "docker",
            "compose",
            "exec",
            self.service_name,
            "find",
            "/usr",
            "-name",
            "exaplus",
            "-type",
            "f",  # only files
            "-executable",  # only executable files
            "-print",  # -print -quit will stop after the result is found
            "-quit",
        ]
        result = subprocess.run(
            find_exaplus, capture_output=True, check=True, text=True
        )
        return result.stdout.strip()

    def _load_data(self, **_: Any) -> None:
        """Load test data into a backend."""
        ddl_file = f"{self.data_volume}/exasol.sql"
        execute_ddl_file = [
            "docker",
            "compose",
            "exec",
            self.service_name,
            self._exaplus(),
            "-c",
            f"{EXASOL_HOST}:{EXASOL_PORT}",
            "-u",
            EXASOL_USER,
            "-p",
            EXASOL_PASS,
            "-f",
            ddl_file,
            "--jdbcparam",
            "validateservercertificate=0",
        ]
        subprocess.check_call(execute_ddl_file)

    def preload(self):
        # copy data files
        super().preload()

        service = self.service_name
        data_volume = self.data_volume
        path = self.script_dir / f"{self.name()}.sql"

        subprocess.check_call(
            [
                "docker",
                "compose",
                "cp",
                f"{path}",
                f"{service}:{data_volume}/{path.name}",
            ]
        )



from __future__ import annotations

import functools
import re
import sys
from collections.abc import Callable as CallableABC
from collections.abc import Sequence
from dataclasses import dataclass
from typing import (  # noqa: UP035
    Annotated,
    Callable,
    Generic,
    List,
    Literal,
    Optional,
    TypeVar,
    Union,
)
from typing import (
    Any as AnyType,
)

import pytest

from ibis.common.annotations import ValidationError
from ibis.common.collections import FrozenDict
from ibis.common.deferred import Call, deferred, var
from ibis.common.graph import Node as GraphNode
from ibis.common.patterns import (
    AllOf,
    Any,
    AnyOf,
    Between,
    CallableWith,
    Capture,
    Check,
    CoercedTo,
    Coercible,
    Contains,
    Custom,
    DictOf,
    EqualTo,
    FrozenDictOf,
    GenericInstanceOf,
    GenericSequenceOf,
    InstanceOf,
    IsIn,
    LazyInstanceOf,
    Length,
    ListOf,
    MappingOf,
    Node,
    NoMatch,
    NoneOf,
    Not,
    Nothing,
    Object,
    Option,
    Pattern,
    PatternList,
    Replace,
    SequenceOf,
    Some,
    SubclassOf,
    TupleOf,
    TypeOf,
    Variable,
    _,
    match,
    pattern,
    replace,
)
from ibis.util import Namespace


class Double(Pattern):
    def match(self, value, *, context):
        return value * 2

    def __eq__(self, other):
        return type(self) == type(other)

    def __hash__(self):
        return hash(type(self))


class Min(Pattern):
    __slots__ = ("min",)

    def __init__(self, min):
        self.min = min

    def match(self, value, context):
        if value >= self.min:
            return value
        else:
            return NoMatch

    def __hash__(self):
        return hash((self.__class__, self.min))

    def __eq__(self, other):
        return self.__class__ == other.__class__ and self.min == other.min


x = Variable("x")
y = Variable("y")
z = Variable("z")


def test_immutability_of_patterns():
    p = InstanceOf(int)
    with pytest.raises(AttributeError):
        p.types = [str]


def test_nothing():
    p = Nothing()
    assert p.match(1, context={}) is NoMatch
    assert p.match(2, context={}) is NoMatch


def test_min():
    p = Min(10)
    assert p.match(10, context={}) == 10
    assert p.match(9, context={}) is NoMatch


def test_double():
    p = Double()
    assert p.match(10, context={}) == 20


def test_any():
    p = Any()
    assert p.match(1, context={}) == 1
    assert p.match("foo", context={}) == "foo"
    assert p.describe() == "matching Any()"


def test_pattern_factory_wraps_variable_with_capture():
    v = var("other")
    p = pattern(v)
    assert p == Capture("other", Any())

    ctx = {}
    assert p.match(10, ctx) == 10
    assert ctx == {"other": 10}


def test_match_on_ellipsis():
    assert match(..., 1) == 1
    assert match(..., [1, 2, 3]) == [1, 2, 3]
    assert match(..., (1, 2, 3)) == (1, 2, 3)


def test_capture():
    ctx = {}

    p = Capture("result", Min(11))
    assert p.match(10, context=ctx) is NoMatch
    assert ctx == {}

    assert p.match(12, context=ctx) == 12
    assert ctx == {"result": 12}


def test_option():
    p = Option(InstanceOf(str))
    assert Option(str) == p
    assert p.match(None, context={}) is None
    assert p.match("foo", context={}) == "foo"
    assert p.match(1, context={}) is NoMatch
    assert p.describe() == "either None or a str"
    assert p.describe(plural=True) == "optional strs"

    p = Option(int, default=-1)
    assert p.match(None, context={}) == -1
    assert p.match(1, context={}) == 1
    assert p.match(1.0, context={}) is NoMatch
    assert p.describe() == "either None or an int"
    assert p.describe(plural=True) == "optional ints"


def test_check():
    def checker(x):
        return x == 10

    p = Check(checker)
    assert p.match(10, context={}) == 10
    assert p.match(11, context={}) is NoMatch
    assert p.describe() == "a value that satisfies checker()"
    assert p.describe(plural=True) == "values that satisfy checker()"


def test_equal_to():
    p = EqualTo(10)
    assert p.match(10, context={}) == 10
    assert p.match(11, context={}) is NoMatch
    assert p.describe() == "10"
    assert p.describe(plural=True) == "10"

    p = EqualTo("10")
    assert p.match(10, context={}) is NoMatch
    assert p.match("10", context={}) == "10"
    assert p.describe() == "'10'"
    assert p.describe(plural=True) == "'10'"


def test_type_of():
    p = TypeOf(int)
    assert p.match(1, context={}) == 1
    assert p.match("foo", context={}) is NoMatch
    assert p.describe() == "exactly an int"
    assert p.describe(plural=True) == "exactly ints"


def test_subclass_of():
    p = SubclassOf(Pattern)
    assert p.match(Double, context={}) == Double
    assert p.match(int, context={}) is NoMatch
    assert p.describe() == "a subclass of Pattern"
    assert p.describe(plural=True) == "subclasses of Pattern"


def test_instance_of():
    p = InstanceOf(int)
    assert p.match(1, context={}) == 1
    assert p.match("foo", context={}) is NoMatch
    assert p.describe() == "an int"
    assert p.describe(plural=True) == "ints"

    p = InstanceOf((int, str))
    assert p.match(1, context={}) == 1
    assert p.match("foo", context={}) == "foo"
    assert p.match(1.0, context={}) is NoMatch
    assert p.describe() == "an int or a str"
    assert p.describe(plural=True) == "ints or strs"

    p = InstanceOf((int, str, float))
    assert p.describe() == "an int, a str or a float"


def test_lazy_instance_of():
    p = LazyInstanceOf("re.Pattern")
    assert p.match(re.compile("foo"), context={}) == re.compile("foo")
    assert p.match("foo", context={}) is NoMatch


T = TypeVar("T", covariant=True)
S = TypeVar("S", covariant=True)


@dataclass
class My(Generic[T, S]):
    a: T
    b: S
    c: str


def test_generic_instance_of_with_covariant_typevar():
    p = Pattern.from_typehint(My[int, AnyType])
    assert p.match(My(1, 2, "3"), context={}) == My(1, 2, "3")
    assert p.describe() == "a My[int, Any]"

    assert match(My[int, AnyType], v := My(1, 2, "3")) == v  # noqa: RUF018
    assert match(My[int, int], v := My(1, 2, "3")) == v  # noqa: RUF018
    assert match(My[int, float], My(1, 2, "3")) is NoMatch
    assert match(My[int, float], v := My(1, 2.0, "3")) == v  # noqa: RUF018


def test_generic_instance_of_disallow_nested_coercion():
    class MyString(str, Coercible):
        @classmethod
        def __coerce__(cls, other):
            return cls(str(other))

    class Box(Generic[T]):
        value: T

    p = Pattern.from_typehint(Box[MyString])
    assert isinstance(p, GenericInstanceOf)
    assert p.origin == Box
    assert p.fields == {"value": InstanceOf(MyString)}


def test_coerced_to():
    class MyInt(int, Coercible):
        @classmethod
        def __coerce__(cls, other):
            return MyInt(MyInt(other) + 1)

    p = CoercedTo(int)
    assert p.match(1, context={}) == 1
    assert p.match("1", context={}) == 1
    with pytest.raises(ValueError):
        p.match("foo", context={})

    p = CoercedTo(MyInt)
    assert p.match(1, context={}) == 2
    assert p.match("1", context={}) == 2
    with pytest.raises(ValueError):
        p.match("foo", context={})


def test_generic_coerced_to():
    class DataType:
        def __eq__(self, other):
            return type(self) == type(other)

    class Integer(DataType):
        pass

    class String(DataType):
        pass

    class DataShape:
        def __eq__(self, other):
            return type(self) == type(other)

    class Scalar(DataShape):
        pass

    class Array(DataShape):
        pass

    class Value(Generic[T, S], Coercible):
        @classmethod
        def __coerce__(cls, value, T=..., S=...):
            return cls(value, Scalar())

        def dtype(self) -> T:
            ...

        def shape(self) -> S:
            ...

    class Literal(Value[T, Scalar]):
        __slots__ = ("_value", "_dtype")

        def __init__(self, value, dtype):
            self._value = value
            self._dtype = dtype

        def dtype(self) -> T:
            return self.dtype

        def shape(self) -> DataShape:
            return Scalar()

        def __eq__(self, other):
            return (
                type(self) == type(other)
                and self._value == other._value
                and self._dtype == other._dtype
            )

    p = Pattern.from_typehint(Literal[String])
    r = p.match("foo", context={})
    assert r == Literal("foo", Scalar())
    expected = "coercible to a <locals>.Literal[<locals>.String]"
    assert p.describe() == expected


def test_not():
    p = Not(InstanceOf(int))
    p1 = ~InstanceOf(int)

    assert p == p1
    assert p.match(1, context={}) is NoMatch
    assert p.match("foo", context={}) == "foo"
    assert p.describe() == "anything except an int"
    assert p.describe(plural=True) == "anything except ints"


def test_any_of():
    p = AnyOf(InstanceOf(int), InstanceOf(str))
    p1 = InstanceOf(int) | InstanceOf(str)

    assert p == p1
    assert p.match(1, context={}) == 1
    assert p.match("foo", context={}) == "foo"
    assert p.match(1.0, context={}) is NoMatch
    assert p.describe() == "an int or a str"
    assert p.describe(plural=True) == "ints or strs"

    p = AnyOf(InstanceOf(int), InstanceOf(str), InstanceOf(float))
    assert p.describe() == "an int, a str or a float"


def test_all_of():
    def negative(x):
        return x < 0

    p = AllOf(InstanceOf(int), Check(negative))
    p1 = InstanceOf(int) & Check(negative)

    assert p == p1
    assert p.match(1, context={}) is NoMatch
    assert p.match(-1, context={}) == -1
    assert p.match(1.0, context={}) is NoMatch
    assert p.describe() == "an int then a value that satisfies negative()"

    p = AllOf(InstanceOf(int), CoercedTo(float), CoercedTo(str))
    assert p.match(1, context={}) == "1.0"
    assert p.match(1.0, context={}) is NoMatch
    assert p.match("1", context={}) is NoMatch
    assert p.describe() == "an int, coercible to a float then coercible to a str"


def test_none_of():
    def negative(x):
        return x < 0

    p = NoneOf(InstanceOf(int), Check(negative))
    assert p.match(1.0, context={}) == 1.0
    assert p.match(-1.0, context={}) is NoMatch
    assert p.match(1, context={}) is NoMatch
    assert p.describe() == "anything except an int or a value that satisfies negative()"


def test_length():
    with pytest.raises(ValueError):
        Length(exactly=3, at_least=3)
    with pytest.raises(ValueError):
        Length(exactly=3, at_most=3)

    p = Length(exactly=3)
    assert p.match([1, 2, 3], context={}) == [1, 2, 3]
    assert p.match([1, 2], context={}) is NoMatch
    assert p.describe() == "with length exactly 3"

    p = Length(at_least=3)
    assert p.match([1, 2, 3], context={}) == [1, 2, 3]
    assert p.match([1, 2], context={}) is NoMatch
    assert p.describe() == "with length at least 3"

    p = Length(at_most=3)
    assert p.match([1, 2, 3], context={}) == [1, 2, 3]
    assert p.match([1, 2, 3, 4], context={}) is NoMatch
    assert p.describe() == "with length at most 3"

    p = Length(at_least=3, at_most=5)
    assert p.match([1, 2], context={}) is NoMatch
    assert p.match([1, 2, 3], context={}) == [1, 2, 3]
    assert p.match([1, 2, 3, 4], context={}) == [1, 2, 3, 4]
    assert p.match([1, 2, 3, 4, 5], context={}) == [1, 2, 3, 4, 5]
    assert p.match([1, 2, 3, 4, 5, 6], context={}) is NoMatch
    assert p.describe() == "with length between 3 and 5"


def test_contains():
    p = Contains(1)
    assert p.match([1, 2, 3], context={}) == [1, 2, 3]
    assert p.match([2, 3], context={}) is NoMatch
    assert p.match({1, 2, 3}, context={}) == {1, 2, 3}
    assert p.match({2, 3}, context={}) is NoMatch
    assert p.describe() == "containing 1"
    assert p.describe(plural=True) == "containing 1"

    p = Contains("1")
    assert p.match([1, 2, 3], context={}) is NoMatch
    assert p.match(["1", 2, 3], context={}) == ["1", 2, 3]
    assert p.match("123", context={}) == "123"
    assert p.describe() == "containing '1'"


def test_isin():
    p = IsIn([1, 2, 3])
    assert p.match(1, context={}) == 1
    assert p.match(4, context={}) is NoMatch
    assert p.describe() == "in {1, 2, 3}"
    assert p.describe(plural=True) == "in {1, 2, 3}"


def test_sequence_of():
    p = SequenceOf(InstanceOf(str), list)
    assert isinstance(p, SequenceOf)
    assert p.match(["foo", "bar"], context={}) == ["foo", "bar"]
    assert p.match([1, 2], context={}) is NoMatch
    assert p.match(1, context={}) is NoMatch
    assert p.match("string", context={}) is NoMatch
    assert p.describe() == "a list of strs"
    assert p.describe(plural=True) == "lists of strs"


def test_generic_sequence_of():
    class MyList(list, Coercible):
        @classmethod
        def __coerce__(cls, value, T=...):
            return cls(value)

    p = GenericSequenceOf(InstanceOf(str), MyList)
    assert p.match(["foo", "bar"], context={}) == MyList(["foo", "bar"])
    assert p.match("string", context={}) is NoMatch

    p = GenericSequenceOf(InstanceOf(str), tuple, at_least=1)
    assert p == GenericSequenceOf(InstanceOf(str), tuple, at_least=1)
    assert p.match(("foo", "bar"), context={}) == ("foo", "bar")
    assert p.match([], context={}) is NoMatch


def test_list_of():
    p = ListOf(InstanceOf(str))
    assert isinstance(p, SequenceOf)
    assert p.match(["foo", "bar"], context={}) == ["foo", "bar"]
    assert p.match([1, 2], context={}) is NoMatch
    assert p.match(1, context={}) is NoMatch
    assert p.describe() == "a list of strs"
    assert p.describe(plural=True) == "lists of strs"


def test_pattern_sequence():
    p = PatternList((InstanceOf(str), InstanceOf(int), InstanceOf(float)))
    assert p.match(("foo", 1, 1.0), context={}) == ["foo", 1, 1.0]
    assert p.match(["foo", 1, 1.0], context={}) == ["foo", 1, 1.0]
    assert p.match(1, context={}) is NoMatch
    assert p.describe() == "a tuple of (a str, an int, a float)"
    assert p.describe(plural=True) == "tuples of (a str, an int, a float)"

    p = PatternList((InstanceOf(str),))
    assert p.match(("foo",), context={}) == ["foo"]
    assert p.match(("foo", "bar"), context={}) is NoMatch


def test_mapping_of():
    p = MappingOf(InstanceOf(str), InstanceOf(int))
    assert p.match({"foo": 1, "bar": 2}, context={}) == {"foo": 1, "bar": 2}
    assert p.match({"foo": 1, "bar": "baz"}, context={}) is NoMatch
    assert p.match(1, context={}) is NoMatch

    p = MappingOf(InstanceOf(str), InstanceOf(str), FrozenDict)
    assert p.match({"foo": "bar"}, context={}) == FrozenDict({"foo": "bar"})
    assert p.match({"foo": 1}, context={}) is NoMatch


class Foo:
    __match_args__ = ("a", "b")

    def __init__(self, a, b):
        self.a = a
        self.b = b

    def __eq__(self, other):
        return type(self) == type(other) and self.a == other.a and self.b == other.b


class Bar:
    __match_args__ = ("c", "d")

    def __init__(self, c, d):
        self.c = c
        self.d = d

    def __eq__(self, other):
        return type(self) == type(other) and self.c == other.c and self.d == other.d


def test_object_pattern():
    p = Object(Foo, 1, b=2)
    o = Foo(1, 2)
    r = match(p, o)
    assert r is o
    assert r == Foo(1, 2)


def test_object_pattern_complex_type():
    p = Object(Not(Foo), 1, 2)
    o = Bar(1, 2)

    # test that the pattern isn't changing the input object if none of
    # its arguments are changed by subpatterns
    assert match(p, o) is o
    assert match(p, Foo(1, 2)) is NoMatch
    assert match(p, Bar(1, 3)) is NoMatch

    p = Object(Not(Foo), 1, b=2)
    assert match(p, Bar(1, 2)) is NoMatch


def test_object_pattern_from_instance_of():
    class MyType:
        __match_args__ = ("a", "b")

        def __init__(self, a, b):
            self.a = a
            self.b = b

    p = pattern(MyType)
    assert p == InstanceOf(MyType)

    p_call = p(1, 2)
    assert p_call == Object(MyType, 1, 2)


def test_object_pattern_from_coerced_to():
    class MyCoercibleType(Coercible):
        __match_args__ = ("a", "b")

        def __init__(self, a, b):
            self.a = a
            self.b = b

        @classmethod
        def __coerce__(cls, other):
            a, b = other
            return cls(a, b)

    p = CoercedTo(MyCoercibleType)
    p_call = p(1, 2)
    assert p_call == Object(MyCoercibleType, 1, 2)


def test_object_pattern_matching_order():
    class Foo:
        __match_args__ = ("a", "b", "c")

        def __init__(self, a, b, c):
            self.a = a
            self.b = b
            self.c = c

        def __eq__(self, other):
            return (
                type(self) == type(other)
                and self.a == other.a
                and self.b == other.b
                and self.c == other.c
            )

    a = var("a")
    p = Object(Foo, a, c=EqualTo(a))

    assert match(p, Foo(1, 2, 3)) is NoMatch
    assert match(p, Foo(1, 2, 1)) == Foo(1, 2, 1)


def test_object_pattern_matching_dictionary_field():
    a = Bar(1, FrozenDict())
    b = Bar(1, {})
    c = Bar(1, None)
    d = Bar(1, {"foo": 1})

    pattern = Object(Bar, 1, d={})
    assert match(pattern, a) is a
    assert match(pattern, b) is b
    assert match(pattern, c) is NoMatch

    pattern = Object(Bar, 1, d=None)
    assert match(pattern, a) is NoMatch
    assert match(pattern, c) is c

    pattern = Object(Bar, 1, d={"foo": 1})
    assert match(pattern, a) is NoMatch
    assert match(pattern, d) is d


def test_object_pattern_requires_its_arguments_to_match():
    class Empty:
        __match_args__ = ()

    msg = "The type to match has fewer `__match_args__`"
    with pytest.raises(ValueError, match=msg):
        Object(Empty, 1)

    # if the type matcher (first argument of Object) receives a generic pattern
    # instead of an explicit type, the validation above cannot occur, so test
    # the the pattern still doesn't match when it requires more positional
    # arguments than the object `__match_args__` has
    pattern = Object(InstanceOf(Empty), var("a"))
    assert match(pattern, Empty()) is NoMatch

    pattern = Object(InstanceOf(Empty), a=var("a"))
    assert match(pattern, Empty()) is NoMatch


def test_callable_with():
    def func(a, b):
        return str(a) + b

    def func_with_args(a, b, *args):
        return sum((a, b) + args)

    def func_with_kwargs(a, b, c=1, **kwargs):
        return str(a) + b + str(c)

    def func_with_optional_keyword_only_kwargs(a, *, c=1):
        return a + c

    def func_with_required_keyword_only_kwargs(*, c):
        return c

    p = CallableWith([InstanceOf(int), InstanceOf(str)])
    assert p.match(10, context={}) is NoMatch

    msg = "Callable has mandatory keyword-only arguments which cannot be specified"
    with pytest.raises(TypeError, match=msg):
        p.match(func_with_required_keyword_only_kwargs, context={})

    # Callable has more positional arguments than expected
    p = CallableWith([InstanceOf(int)] * 2)
    assert p.match(func_with_kwargs, context={}).__wrapped__ is func_with_kwargs

    # Callable has less positional arguments than expected
    p = CallableWith([InstanceOf(int)] * 4)
    assert p.match(func_with_kwargs, context={}) is NoMatch

    p = CallableWith([InstanceOf(int)] * 4, InstanceOf(int))
    wrapped = p.match(func_with_args, context={})
    assert wrapped(1, 2, 3, 4) == 10

    p = CallableWith([InstanceOf(int), InstanceOf(str)], InstanceOf(str))
    wrapped = p.match(func, context={})
    assert wrapped(1, "st") == "1st"

    with pytest.raises(ValidationError):
        wrapped(1, 2)

    p = CallableWith([InstanceOf(int)])
    wrapped = p.match(func_with_optional_keyword_only_kwargs, context={})
    assert wrapped(1) == 2


def test_callable_with_default_arguments():
    def f(a: int, b: str, c: str):
        return a + int(b) + int(c)

    def g(a: int, b: str, c: str = "0"):
        return a + int(b) + int(c)

    h = functools.partial(f, c="0")

    p = Pattern.from_typehint(Callable[[int, str], int])
    assert p.match(f, {}) is NoMatch
    assert p.match(g, {}).__wrapped__ == g
    assert p.match(h, {}).__wrapped__ == h


def test_pattern_list():
    p = PatternList([1, 2, InstanceOf(int), Some(...)])
    assert p.match([1, 2, 3, 4, 5], context={}) == [1, 2, 3, 4, 5]
    assert p.match([1, 2, 3, 4, 5, 6], context={}) == [1, 2, 3, 4, 5, 6]
    assert p.match([1, 2, 3, 4], context={}) == [1, 2, 3, 4]
    assert p.match([1, 2, "3", 4], context={}) is NoMatch

    # subpattern is a simple pattern
    p = PatternList([1, 2, CoercedTo(int), Some(...)])
    assert p.match([1, 2, 3.0, 4.0, 5.0], context={}) == [1, 2, 3, 4.0, 5.0]

    # subpattern is a sequence
    p = PatternList([1, 2, 3, Some(CoercedTo(int), at_least=1)])
    assert p.match([1, 2, 3, 4.0, 5.0], context={}) == [1, 2, 3, 4, 5]


def test_pattern_list_from_tuple_typehint():
    p = Pattern.from_typehint(tuple[str, int, float])
    assert p == PatternList(
        [InstanceOf(str), InstanceOf(int), InstanceOf(float)], type=tuple
    )
    assert p.match(["foo", 1, 2.0], context={}) == ("foo", 1, 2.0)
    assert p.match(("foo", 1, 2.0), context={}) == ("foo", 1, 2.0)
    assert p.match(["foo", 1], context={}) is NoMatch
    assert p.match(["foo", 1, 2.0, 3.0], context={}) is NoMatch

    class MyTuple(tuple):
        pass

    p = Pattern.from_typehint(MyTuple[int, bool])
    assert p == PatternList([InstanceOf(int), InstanceOf(bool)], type=MyTuple)
    assert p.match([1, True], context={}) == MyTuple([1, True])
    assert p.match(MyTuple([1, True]), context={}) == MyTuple([1, True])
    assert p.match([1, 2], context={}) is NoMatch


def test_pattern_list_unpack():
    integer = pattern(int)
    floating = pattern(float)

    assert match([1, 2, *floating], [1, 2, 3]) is NoMatch
    assert match([1, 2, *floating], [1, 2, 3.0]) == [1, 2, 3.0]
    assert match([1, 2, *floating], [1, 2, 3.0, 4.0]) == [1, 2, 3.0, 4.0]
    assert match([1, *floating, *integer], [1, 2.0, 3.0, 4]) == [1, 2.0, 3.0, 4]
    assert match([1, *floating, *integer], [1, 2.0, 3.0, 4, 5]) == [
        1,
        2.0,
        3.0,
        4,
        5,
    ]
    assert match([1, *floating, *integer], [1, 2.0, 3, 4.0]) is NoMatch


def test_matching():
    assert match("foo", "foo") == "foo"
    assert match("foo", "bar") is NoMatch

    assert match(InstanceOf(int), 1) == 1
    assert match(InstanceOf(int), "foo") is NoMatch

    assert Capture("pi", InstanceOf(float)) == "pi" @ InstanceOf(float)
    assert Capture("pi", InstanceOf(float)) == "pi" @ InstanceOf(float)

    assert match(Capture("pi", InstanceOf(float)), 3.14, ctx := {}) == 3.14  # noqa: RUF018
    assert ctx == {"pi": 3.14}
    assert match("pi" @ InstanceOf(float), 3.14, ctx := {}) == 3.14  # noqa: RUF018
    assert ctx == {"pi": 3.14}

    assert match("pi" @ InstanceOf(float), 3.14, ctx := {}) == 3.14  # noqa: RUF018
    assert ctx == {"pi": 3.14}

    assert match(InstanceOf(int) | InstanceOf(float), 3) == 3
    assert match(InstanceOf(object) & InstanceOf(float), 3.14) == 3.14


def test_replace_passes_matched_value_as_underscore():
    class MyInt:
        def __init__(self, value):
            self.value = value

        def __eq__(self, other):
            return self.value == other.value

    p = InstanceOf(int) >> Call(MyInt, value=_)
    assert p.match(1, context={}) == MyInt(1)


def test_replace_in_nested_object_pattern():
    # simple example using reference to replace a value
    b = Variable("b")
    p = Object(Foo, 1, b=Replace(Any(), b))
    f = p.match(Foo(1, 2), {"b": 3})
    assert f.a == 1
    assert f.b == 3

    # nested example using reference to replace a value
    d = Variable("d")
    p = Object(Foo, 1, b=Object(Bar, 2, d=Replace(Any(), d)))
    g = p.match(Foo(1, Bar(2, 3)), {"d": 4})
    assert g.b.c == 2
    assert g.b.d == 4

    # nested example using reference to replace a value with a captured value
    p = Object(
        Foo,
        1,
        b=Replace(Object(Bar, 2, d="d" @ Any()), lambda _, d: Foo(-1, b=d)),
    )
    h = p.match(Foo(1, Bar(2, 3)), {})
    assert isinstance(h, Foo)
    assert h.a == 1
    assert isinstance(h.b, Foo)
    assert h.b.b == 3

    # same example with more syntactic sugar
    o = Namespace(pattern, module=__name__)
    c = Namespace(deferred, module=__name__)

    d = Variable("d")
    p = o.Foo(1, b=o.Bar(2, d=d @ Any()) >> c.Foo(-1, b=d))
    h1 = p.match(Foo(1, Bar(2, 3)), {})
    assert isinstance(h1, Foo)
    assert h1.a == 1
    assert isinstance(h1.b, Foo)
    assert h1.b.b == 3


def test_replace_decorator():
    @replace(int)
    def sub(_):
        return _ - 1

    assert match(sub, 1) == 0
    assert match(sub, 2) == 1


def test_replace_using_deferred():
    p = Namespace(pattern, module=__name__)
    d = Namespace(deferred, module=__name__)

    x = var("x")
    y = var("y")

    pat = p.Foo(x, b=y) >> d.Foo(x, b=y)
    assert match(pat, Foo(1, 2)) == Foo(1, 2)

    pat = p.Foo(x, b=y) >> d.Foo(x, b=(y + 1) * x)
    assert match(pat, Foo(2, 3)) == Foo(2, 8)

    pat = p.Foo(x, y @ p.Bar) >> d.Foo(x, b=y.c + y.d)
    assert match(pat, Foo(1, Bar(2, 3))) == Foo(1, 5)


def test_matching_sequence_pattern():
    assert match([], []) == []
    assert match([], [1]) is NoMatch

    assert match([1, 2, 3, 4, Some(...)], list(range(1, 9))) == list(range(1, 9))
    assert match([1, 2, 3, 4, Some(...)], list(range(1, 3))) is NoMatch
    assert match([1, 2, 3, 4, Some(...)], list(range(1, 5))) == list(range(1, 5))
    assert match([1, 2, 3, 4, Some(...)], list(range(1, 6))) == list(range(1, 6))

    assert match([Some(...), 3, 4], list(range(5))) == list(range(5))
    assert match([Some(...), 3, 4], list(range(3))) is NoMatch

    assert match([0, 1, Some(...), 4], list(range(5))) == list(range(5))
    assert match([0, 1, Some(...), 4], list(range(4))) is NoMatch

    assert match([Some(...)], list(range(5))) == list(range(5))
    assert match([Some(...), 2, 3, 4, Some(...)], list(range(8))) == list(range(8))


def test_matching_sequence_pattern_keeps_original_type():
    assert match([1, 2, 3, 4, Some(...)], tuple(range(1, 9))) == list(range(1, 9))
    assert match((1, 2, 3, Some(...)), [1, 2, 3, 4, 5]) == (1, 2, 3, 4, 5)


def test_matching_sequence_with_captures():
    v = list(range(1, 9))
    assert match([1, 2, 3, 4, Some(...)], v) == v
    assert match([1, 2, 3, 4, "rest" @ Some(...)], v, ctx := {}) == v  # noqa: RUF018
    assert ctx == {"rest": [5, 6, 7, 8]}

    v = list(range(5))
    assert match([0, 1, x @ Some(...), 4], v, ctx := {}) == v  # noqa: RUF018
    assert ctx == {"x": [2, 3]}
    assert match([0, 1, "var" @ Some(...), 4], v, ctx := {}) == v  # noqa: RUF018
    assert ctx == {"var": [2, 3]}

    p = [
        0,
        1,
        "ints" @ Some(int),
        Some("last_float" @ InstanceOf(float)),
        6,
    ]
    v = [0, 1, 2, 3, 4.0, 5.0, 6]
    assert match(p, v, ctx := {}) == v  # noqa: RUF018
    assert ctx == {"ints": [2, 3], "last_float": 5.0}


def test_matching_sequence_remaining():
    three = [1, 2, 3]
    four = [1, 2, 3, 4]
    five = [1, 2, 3, 4, 5]

    assert match([1, 2, 3, Some(int, at_least=1)], four) == four
    assert match([1, 2, 3, Some(int, at_least=1)], three) is NoMatch
    assert match([1, 2, 3, Some(int)], three) == three
    assert match([1, 2, 3, Some(int, at_most=1)], three) == three
    assert match([1, 2, 3, Some(InstanceOf(int) & Between(0, 10))], five) == five
    assert match([1, 2, 3, Some(InstanceOf(int) & Between(0, 4))], five) is NoMatch
    assert match([1, 2, 3, Some(int, at_least=2)], four) is NoMatch
    assert match([1, 2, 3, "res" @ Some(int, at_least=2)], five, ctx := {}) == five  # noqa: RUF018
    assert ctx == {"res": [4, 5]}


def test_matching_sequence_complicated():
    pat = [
        1,
        "a" @ Some(InstanceOf(int) & Check(lambda x: x < 10)),
        4,
        "b" @ Some(...),
        8,
        9,
    ]
    expected = {
        "a": [2, 3],
        "b": [5, 6, 7],
    }
    assert match(pat, range(1, 10), ctx := {}) == list(range(1, 10))  # noqa: RUF018
    assert ctx == expected

    pat = [1, 2, Capture("remaining", Some(...))]
    expected = {"remaining": [3, 4, 5, 6, 7, 8, 9]}
    assert match(pat, range(1, 10), ctx := {}) == list(range(1, 10))  # noqa: RUF018
    assert ctx == expected

    v = [0, [1, 2, "3"], [1, 2, "4"], 3]
    assert match([0, Some([1, 2, str]), 3], v) == v


def test_pattern_sequence_with_nested_some():
    ctx = {}
    res = match([0, "subseq" @ Some(1, 2), 3], [0, 1, 2, 1, 2, 3], ctx)
    assert res == [0, 1, 2, 1, 2, 3]
    assert ctx == {"subseq": [1, 2, 1, 2]}

    assert match([0, Some(1), 2, 3], [0, 2, 3]) == [0, 2, 3]
    assert match([0, Some(1, at_least=1), 2, 3], [0, 2, 3]) is NoMatch
    assert match([0, Some(1, at_least=1), 2, 3], [0, 1, 2, 3]) == [0, 1, 2, 3]
    assert match([0, Some(1, at_least=2), 2, 3], [0, 1, 2, 3]) is NoMatch
    assert match([0, Some(1, at_least=2), 2, 3], [0, 1, 1, 2, 3]) == [0, 1, 1, 2, 3]
    assert match([0, Some(1, at_most=2), 2, 3], [0, 1, 1, 2, 3]) == [0, 1, 1, 2, 3]
    assert match([0, Some(1, at_most=1), 2, 3], [0, 1, 1, 2, 3]) is NoMatch
    assert match([0, Some(1, exactly=1), 2, 3], [0, 2, 3]) is NoMatch
    assert match([0, Some(1, exactly=1), 2, 3], [0, 1, 2, 3]) == [0, 1, 2, 3]
    assert match([0, Some(1, exactly=0), 2, 3], [0, 2, 3]) == [0, 2, 3]
    assert match([0, Some(1, exactly=0), 2, 3], [0, 1, 2, 3]) is NoMatch

    assert match([0, Some(1, Some(2)), 3], [0, 3]) == [0, 3]
    assert match([0, Some(1, Some(2)), 3], [0, 1, 3]) == [0, 1, 3]
    assert match([0, Some(1, Some(2)), 3], [0, 1, 2, 3]) == [0, 1, 2, 3]
    assert match([0, Some(1, Some(2)), 3], [0, 1, 2, 2, 3]) == [0, 1, 2, 2, 3]
    assert match([0, Some(1, Some(2)), 3], [0, 1, 2, 2, 2, 3]) == [0, 1, 2, 2, 2, 3]
    assert match([0, Some(1, Some(2)), 3], [0, 1, 2, 1, 2, 2, 3]) == [
        0,
        1,
        2,
        1,
        2,
        2,
        3,
    ]
    assert match([0, Some(1, Some(2), at_least=1), 3], [0, 1, 2, 3]) == [0, 1, 2, 3]
    assert match([0, Some(1, Some(2), at_least=1), 3], [0, 1, 3]) == [0, 1, 3]
    assert match([0, Some(1, Some(2, at_least=2), at_least=1), 3], [0, 1, 3]) is NoMatch
    assert (
        match([0, Some(1, Some(2, at_least=2), at_least=1), 3], [0, 1, 2, 3]) is NoMatch
    )
    assert match([0, Some(1, Some(2, at_least=2), at_least=1), 3], [0, 1, 2, 2, 3]) == [
        0,
        1,
        2,
        2,
        3,
    ]


@pytest.mark.parametrize(
    ("pattern", "value", "expected"),
    [
        (InstanceOf(bool), True, True),
        (InstanceOf(str), "foo", "foo"),
        (InstanceOf(int), 8, 8),
        (InstanceOf(int), 1, 1),
        (InstanceOf(float), 1.0, 1.0),
        (IsIn({"a", "b"}), "a", "a"),
        (IsIn({"a": 1, "b": 2}), "a", "a"),
        (IsIn(["a", "b"]), "a", "a"),
        (IsIn(("a", "b")), "b", "b"),
        (IsIn({"a", "b", "c"}), "c", "c"),
        (TupleOf(InstanceOf(int)), (1, 2, 3), (1, 2, 3)),
        (PatternList((InstanceOf(int), InstanceOf(str))), (1, "a"), [1, "a"]),
        (ListOf(InstanceOf(str)), ["a", "b"], ["a", "b"]),
        (AnyOf(InstanceOf(str), InstanceOf(int)), "foo", "foo"),
        (AnyOf(InstanceOf(str), InstanceOf(int)), 7, 7),
        (
            AllOf(InstanceOf(int), Check(lambda v: v >= 3), Check(lambda v: v >= 8)),
            10,
            10,
        ),
        (
            MappingOf(InstanceOf(str), InstanceOf(int)),
            {"a": 1, "b": 2},
            {"a": 1, "b": 2},
        ),
    ],
)
def test_various_patterns(pattern, value, expected):
    assert pattern.match(value, context={}) == expected


@pytest.mark.parametrize(
    ("pattern", "value"),
    [
        (InstanceOf(bool), "foo"),
        (InstanceOf(str), True),
        (InstanceOf(int), 8.1),
        (Min(3), 2),
        (InstanceOf(int), None),
        (InstanceOf(float), 1),
        (IsIn(["a", "b"]), "c"),
        (IsIn({"a", "b"}), "c"),
        (IsIn({"a": 1, "b": 2}), "d"),
        (TupleOf(InstanceOf(int)), (1, 2.0, 3)),
        (ListOf(InstanceOf(str)), ["a", "b", None]),
        (AnyOf(InstanceOf(str), Min(4)), 3.14),
        (AnyOf(InstanceOf(str), Min(10)), 9),
        (AllOf(InstanceOf(int), Min(3), Min(8)), 7),
        (DictOf(InstanceOf(int), InstanceOf(str)), {"a": 1, "b": 2}),
    ],
)
def test_various_not_matching_patterns(pattern, value):
    assert pattern.match(value, context={}) is NoMatch


@pattern
def endswith_d(s, ctx):
    if not s.endswith("d"):
        return NoMatch
    return s


def test_pattern_decorator():
    assert endswith_d.match("abcd", context={}) == "abcd"
    assert endswith_d.match("abc", context={}) is NoMatch


@pytest.mark.parametrize(
    ("annot", "expected"),
    [
        (int, InstanceOf(int)),
        (str, InstanceOf(str)),
        (bool, InstanceOf(bool)),
        (Optional[int], Option(InstanceOf(int))),
        (Optional[Union[str, int]], Option(AnyOf(InstanceOf(str), InstanceOf(int)))),
        (Union[int, str], AnyOf(InstanceOf(int), InstanceOf(str))),
        (Annotated[int, Min(3)], AllOf(InstanceOf(int), Min(3))),
        (list[int], SequenceOf(InstanceOf(int), list)),
        (
            tuple[int, float, str],
            PatternList(
                (InstanceOf(int), InstanceOf(float), InstanceOf(str)), type=tuple
            ),
        ),
        (tuple[int, ...], TupleOf(InstanceOf(int))),
        (
            dict[str, float],
            DictOf(InstanceOf(str), InstanceOf(float)),
        ),
        (FrozenDict[str, int], FrozenDictOf(InstanceOf(str), InstanceOf(int))),
        (Literal["alpha", "beta", "gamma"], IsIn(("alpha", "beta", "gamma"))),
        (
            Callable[[str, int], str],
            CallableWith((InstanceOf(str), InstanceOf(int)), InstanceOf(str)),
        ),
        (Callable, InstanceOf(CallableABC)),
    ],
)
def test_pattern_from_typehint(annot, expected):
    assert Pattern.from_typehint(annot) == expected


@pytest.mark.skipif(sys.version_info < (3, 10), reason="requires python3.10 or higher")
def test_pattern_from_typehint_uniontype():
    # uniontype marks `type1 | type2` annotations and it's different from
    # Union[type1, type2]
    validator = Pattern.from_typehint(str | int | float)
    assert validator == AnyOf(InstanceOf(str), InstanceOf(int), InstanceOf(float))


def test_pattern_from_typehint_disable_coercion():
    class MyFloat(float, Coercible):
        @classmethod
        def __coerce__(cls, obj):
            return cls(float(obj))

    p = Pattern.from_typehint(MyFloat, allow_coercion=True)
    assert isinstance(p, CoercedTo)

    p = Pattern.from_typehint(MyFloat, allow_coercion=False)
    assert isinstance(p, InstanceOf)


class PlusOne(Coercible):
    __slots__ = ("value",)

    def __init__(self, value):
        self.value = value

    @classmethod
    def __coerce__(cls, obj):
        return cls(obj + 1)

    def __eq__(self, other):
        return type(self) == type(other) and self.value == other.value


class PlusOneRaise(PlusOne):
    @classmethod
    def __coerce__(cls, obj):
        if isinstance(obj, cls):
            return obj
        else:
            raise ValueError("raise on coercion")


class PlusOneChild(PlusOne):
    pass


class PlusTwo(PlusOne):
    @classmethod
    def __coerce__(cls, obj):
        return obj + 2


def test_pattern_from_coercible_protocol():
    s = Pattern.from_typehint(PlusOne)
    assert s.match(1, context={}) == PlusOne(2)
    assert s.match(10, context={}) == PlusOne(11)


def test_pattern_coercible_bypass_coercion():
    s = Pattern.from_typehint(PlusOneRaise)
    # bypass coercion since it's already an instance of SomethingRaise
    assert s.match(PlusOneRaise(10), context={}) == PlusOneRaise(10)
    # but actually call __coerce__ if it's not an instance
    with pytest.raises(ValueError, match="raise on coercion"):
        s.match(10, context={})


def test_pattern_coercible_checks_type():
    s = Pattern.from_typehint(PlusOneChild)
    v = Pattern.from_typehint(PlusTwo)

    assert s.match(1, context={}) == PlusOneChild(2)

    assert PlusTwo.__coerce__(1) == 3
    assert v.match(1, context={}) is NoMatch


class DoubledList(Coercible, list[T]):
    @classmethod
    def __coerce__(cls, obj):
        return cls(list(obj) * 2)


def test_pattern_coercible_sequence_type():
    s = Pattern.from_typehint(Sequence[PlusOne])
    with pytest.raises(TypeError, match=r"Sequence\(\) takes no arguments"):
        s.match([1, 2, 3], context={})

    s = Pattern.from_typehint(list[PlusOne])
    assert s == SequenceOf(CoercedTo(PlusOne), type=list)
    assert s.match([1, 2, 3], context={}) == [PlusOne(2), PlusOne(3), PlusOne(4)]

    s = Pattern.from_typehint(tuple[PlusOne, ...])
    assert s == TupleOf(CoercedTo(PlusOne))
    assert s.match([1, 2, 3], context={}) == (PlusOne(2), PlusOne(3), PlusOne(4))

    s = Pattern.from_typehint(DoubledList[PlusOne])
    assert s == GenericSequenceOf(CoercedTo(PlusOne), type=DoubledList)
    assert s.match([1, 2, 3], context={}) == DoubledList(
        [PlusOne(2), PlusOne(3), PlusOne(4), PlusOne(2), PlusOne(3), PlusOne(4)]
    )


def test_pattern_function():
    class MyNegativeInt(int, Coercible):
        @classmethod
        def __coerce__(cls, other):
            return cls(-int(other))

    class Box(Generic[T]):
        value: T

    def f(x):
        return x > 0

    # ... is treated the same as Any()
    assert pattern(...) == Any()
    assert pattern(Any()) == Any()
    assert pattern(True) == EqualTo(True)

    # plain types are converted to InstanceOf patterns
    assert pattern(int) == InstanceOf(int)
    # no matter whether the type implements the coercible protocol or not
    assert pattern(MyNegativeInt) == InstanceOf(MyNegativeInt)

    # generic types are converted to GenericInstanceOf patterns
    assert pattern(Box[int]) == GenericInstanceOf(Box[int])
    # no matter whethwe the origin type implements the coercible protocol or not
    assert pattern(Box[MyNegativeInt]) == GenericInstanceOf(Box[MyNegativeInt])

    # sequence typehints are converted to the appropriate sequence checkers
    assert pattern(List[int]) == ListOf(InstanceOf(int))  # noqa: UP006

    # spelled out sequences construct a more advanced pattern sequence
    assert pattern([int, str, 1]) == PatternList(
        [InstanceOf(int), InstanceOf(str), EqualTo(1)]
    )

    # matching deferred to user defined functions
    assert pattern(f) == Custom(f)

    # matching mapping values
    assert pattern({"a": 1, "b": 2}) == EqualTo(FrozenDict({"a": 1, "b": 2}))


class Term(GraphNode):
    def __eq__(self, other):
        return type(self) is type(other) and self.__args__ == other.__args__

    def __hash__(self):
        return hash((self.__class__, self.__args__))

    def __repr__(self):
        return f"{self.__class__.__name__}({', '.join(map(repr, self.__args__))})"


class Lit(Term):
    __slots__ = ("value",)
    __argnames__ = ("value",)
    __match_args__ = ("value",)

    def __init__(self, value):
        self.value = value

    @property
    def __args__(self):
        return (self.value,)


class Binary(Term):
    __slots__ = ("left", "right")
    __argnames__ = ("left", "right")
    __match_args__ = ("left", "right")

    def __init__(self, left, right):
        self.left = left
        self.right = right

    @property
    def __args__(self):
        return (self.left, self.right)


class Add(Binary):
    pass


class Mul(Binary):
    pass


one = Lit(1)
two = Mul(Lit(2), one)

three = Add(one, two)
six = Mul(two, three)
seven = Add(one, six)
fourteen = Add(seven, seven)


def test_node():
    pat = Node(
        InstanceOf(Add),
        each_arg=Replace(Object(Lit, value=Capture("v")), lambda _, v: Lit(v + 100)),
    )
    result = six.replace(pat)
    assert result == Mul(two, Add(Lit(101), two))



from __future__ import annotations

import collections
import decimal
from typing import TYPE_CHECKING, Union

from ibis.common.dispatch import Dispatched, lazy_singledispatch

# ruff: noqa: F811
if TYPE_CHECKING:
    import pandas as pd
    import pyarrow as pa


def test_lazy_singledispatch():
    @lazy_singledispatch
    def foo(x):
        """A docstring."""
        return "base result"

    @foo.register(int)
    def _(x):
        return x + 1

    @foo.register(float)
    def _(x):
        return x - 1

    @foo.register((tuple, list))
    def _(x):
        return tuple(foo(i) for i in x)

    class Bar:
        pass

    assert foo.__name__ == "foo"
    assert foo.__doc__ == "A docstring."

    assert foo(1) == 2
    assert foo(1.0) == 0.0
    assert foo((1, 2.0)) == (2, 1.0)
    assert foo([1, 2.0]) == (2, 1.0)
    assert foo("a string") == "base result"


def test_lazy_singledispatch_extra_args():
    @lazy_singledispatch
    def foo(*args, **kwargs):
        pass

    @foo.register(int)
    def _(a, b, c=2):
        return a + b + c

    @foo.register(float)
    def _(a, b, c=2):
        return a - b - c

    assert foo(1, 2) == 5
    assert foo(1.0, 2) == -3.0
    assert foo(1, 2, c=3) == 6


def test_lazy_singledispatch_lazy():
    @lazy_singledispatch
    def foo(a):
        return a

    @foo.register("decimal.Decimal")
    def inc(a):
        return a + 1

    assert foo(1) == 1
    assert foo(decimal.Decimal(1)) == decimal.Decimal(2)


def test_lazy_singledispatch_lazy_walks_mro():
    class Subclass(decimal.Decimal):
        pass

    class Subclass2(decimal.Decimal):
        pass

    @lazy_singledispatch
    def foo(a):
        return "base call"

    @foo.register(Subclass2)
    def _(a):
        return "eager call"

    @foo.register("decimal.Decimal")
    def _(a):
        return "lazy call"

    assert foo(1) == "base call"
    assert foo(Subclass2(1)) == "eager call"
    assert foo(Subclass(1)) == "lazy call"
    # not overwritten by lazy loader
    assert foo(Subclass2(1)) == "eager call"


def test_lazy_singledispatch_abc():
    class mydict(dict):
        pass

    @lazy_singledispatch
    def foo(a):
        return "base"

    @foo.register(collections.abc.Mapping)
    def _(a):
        return "mapping"

    @foo.register(mydict)
    def _(a):
        return "mydict"

    @foo.register(collections.abc.Callable)
    def _(a):
        return "callable"

    assert foo(1) == "base"
    assert foo({}) == "mapping"
    assert foo(mydict()) == "mydict"  # concrete takes precedence
    assert foo(sum) == "callable"


class A:
    pass


class B:
    pass


class Visitor(Dispatched):
    def a(self):
        return "a"

    def b(self, x: int):
        return "b_int"

    def b(self, x: str):
        return "b_str"

    def b(self, x: Union[A, B]):
        return "b_union"

    @classmethod
    def c(cls, x: int, **kwargs):
        return "c_int"

    @classmethod
    def c(cls, x: str, a=0, b=1):
        return "c_str"

    def d(self, x: int):
        return "d_int"

    def d(self, x: str):
        return "d_str"

    @staticmethod
    def e(x: int):
        return "e_int"

    @staticmethod
    def e(x: str):
        return "e_str"

    def f(self, df: dict):
        return "f_dict"

    def f(self, df: pd.DataFrame):
        return "f_pandas"

    def f(self, df: pa.Table):
        return "f_pyarrow"


class Subvisitor(Visitor):
    def b(self, x):
        return super().b(x)

    def b(self, x: float):
        return "b_float"

    @classmethod
    def c(cls, x):
        return super().c(x)

    @classmethod
    def c(cls, s: float):
        return "c_float"


def test_dispatched():
    v = Visitor()
    assert v.a() == "a"
    assert v.b(1) == "b_int"
    assert v.b("1") == "b_str"
    assert v.b(A()) == "b_union"
    assert v.b(B()) == "b_union"
    assert v.d(1) == "d_int"
    assert v.d("1") == "d_str"

    w = Subvisitor()
    assert w.b(1) == "b_int"
    assert w.b(1.1) == "b_float"

    assert Visitor.c(1, a=0, b=0) == "c_int"
    assert Visitor.c("1") == "c_str"

    assert Visitor.e("1") == "e_str"
    assert Visitor.e(1) == "e_int"

    assert Subvisitor.c(1) == "c_int"
    assert Subvisitor.c(1.1) == "c_float"

    assert Subvisitor.e(1) == "e_int"


def test_dispatched_lazy():
    import pyarrow as pa

    empty_pyarrow_table = pa.Table.from_arrays([])
    empty_pandas_table = empty_pyarrow_table.to_pandas()

    v = Visitor()
    assert v.f({}) == "f_dict"
    assert v.f(empty_pyarrow_table) == "f_pyarrow"
    assert v.f(empty_pandas_table) == "f_pandas"
